"""
Short summary

Extended summary

See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

import logging

import numpy as np

import ccdproc

from ..test_constants import TEST_WL_IDX


## This and other new functions in this repo should take and return an asdf tree for consistency
## with existing ones.
def flat_correct(data, flats):
    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")

    corrected_frames = []
    ## OPTIMISE: remove loop
    for i, fileobj in enumerate(data):
        ## REFACTOR: file loading
        dataframe = ccdproc.CCDData.read(fileobj.fileuri, format="fits", unit="adu")
        newpath = fileobj.fileuri.replace("dark_corrected_", "flat_corrected_")
        if isinstance(flats, list):
            flat = flats[i]
        elif isinstance(flats, ccdproc.ccddata.CCDData):
            flat = flats
        logger.debug(f"{np.nanmin(flat.data)}")
        dataframe = ccdproc.flat_correct(dataframe, flat, norm_value=1)
        dataframe.write(newpath)
        corrected_frames.append(newpath)

    return corrected_frames



"""
Initial demonstration pipeline.

Mostly for development purposes but also to show a representative example of how the VTF calibration
will run.
"""

import glob
import datetime as dt
from os.path import join
from commands.align_frames import align_frames
from commands.dark_correct import correct_darks
from commands.flat_correct import correct_flats
from commands.reduce_darks import reduce_darks
from commands.reduce_flats import reduce_flats
from commands.reduce_targets import reduce_targets

from airflow import DAG
from airflow.operators.python_operator import PythonOperator
from preprocess_data import process

DATA = join("/", "home", "drew", "DKIST", "vtfcal", "vtfcal", "data")

default_args = {"owner": "airflow", "depends_on_past": False, "start_date": dt.datetime.now()}

dag = DAG("pipeline-demo", default_args=default_args, schedule_interval="@once")
data_tree = join(DATA, "{}_data_tree")


def null():
    """Null function so I've got a thing to pass to bits of the pipeline I haven't written yet"""
    return


preprocess = PythonOperator(
    task_id="Pre-process-data", python_callable=process, op_args=["00000000"], dag=dag
)

"""
Data reduction tasks
"""
# Dark reduction
# --------------
bb_dark_reduction = PythonOperator(
    task_id="Reduce-broadband-darks",
    python_callable=reduce_darks,
    op_args=[data_tree.format("bb")],
    dag=dag,
)

nb1_dark_reduction = PythonOperator(
    task_id="Reduce-narrowband-1-darks",
    python_callable=reduce_darks,
    op_args=[data_tree.format("nb1")],
    dag=dag,
)

nb2_dark_reduction = PythonOperator(
    task_id="Reduce-narrowband-2-darks",
    python_callable=reduce_darks,
    op_args=[data_tree.format("nb2")],
    dag=dag,
)

# Flat reduction
# --------------
bb_flat_reduction = PythonOperator(
    task_id="Reduce-broadband-flats",
    python_callable=reduce_flats,
    op_args=[data_tree.format("bb")],
    dag=dag,
)
bb_dark_reduction >> bb_flat_reduction

nb1_flat_reduction = PythonOperator(
    task_id="Reduce-narrowband-1-flats",
    python_callable=reduce_flats,
    op_args=[data_tree.format("nb1")],
    dag=dag,
)
nb1_dark_reduction >> nb1_flat_reduction

nb2_flat_reduction = PythonOperator(
    task_id="Reduce-narrowband-2-flats",
    python_callable=reduce_flats,
    op_args=[data_tree.format("nb2")],
    dag=dag,
)
nb2_dark_reduction >> nb2_flat_reduction

# Target reduction
# ----------------
target_reduction = PythonOperator(
    task_id="Correct-and-reduce-targets",
    python_callable=reduce_targets,
    op_args=[glob.glob(data_tree.format("*"))],
    dag=dag,
)

# Prefilter reduction
# -------------------
prefil_red = PythonOperator(task_id="Reduce-prefilter", python_callable=null, dag=dag)

# Polarimetric calibration
# ------------------------
polcal = PythonOperator(task_id="Polarisation-calibration", python_callable=null, dag=dag)


"""
Data correction tasks
"""
dark_correction_bb = PythonOperator(
    task_id="Dark-correct-broadband-data",
    python_callable=correct_darks,
    op_args=[data_tree.format("bb")],
    dag=dag,
)

dark_correction_nb1 = PythonOperator(
    task_id="Dark-correct-narrowband1-data",
    python_callable=correct_darks,
    op_args=[data_tree.format("nb1")],
    dag=dag,
)

dark_correction_nb2 = PythonOperator(
    task_id="Dark-correct-narrowband2-data",
    python_callable=correct_darks,
    op_args=[data_tree.format("nb2")],
    dag=dag,
)

bb_dark_reduction >> dark_correction_bb
nb1_dark_reduction >> dark_correction_nb1
nb2_dark_reduction >> dark_correction_nb2

flat_correction_bb = PythonOperator(
    task_id="Flat-correct-broadband-data",
    python_callable=correct_flats,
    op_args=[data_tree.format("bb")],
    dag=dag,
)

flat_correction_nb1 = PythonOperator(
    task_id="Flat-correct-narrowband1-data",
    python_callable=correct_flats,
    op_args=[data_tree.format("nb1")],
    dag=dag,
)

flat_correction_nb2 = PythonOperator(
    task_id="Flat-correct-narrowband2-data",
    python_callable=correct_flats,
    op_args=[data_tree.format("nb2")],
    dag=dag,
)

(bb_flat_reduction, dark_correction_bb) >> flat_correction_bb
(nb1_flat_reduction, dark_correction_nb1) >> flat_correction_nb1
(nb2_flat_reduction, dark_correction_nb2) >> flat_correction_nb2

"""
Reconstruction tasks
"""
data_alignment1 = PythonOperator(
    task_id="Align-narrowband1-data",
    python_callable=align_frames,
    op_args=[data_tree.format("nb1")],
    dag=dag,
)

data_alignment2 = PythonOperator(
    task_id="Align-narrowband2-data",
    python_callable=align_frames,
    op_args=[data_tree.format("nb2")],
    dag=dag,
)

(target_reduction, flat_correction_nb1) >> data_alignment1
(target_reduction, flat_correction_nb2) >> data_alignment2

recon_bb = PythonOperator(task_id="reconstruct-broadband", python_callable=null, dag=dag)

recon_nb1 = PythonOperator(task_id="reconstruct-narrowband-1", python_callable=null, dag=dag)

recon_nb2 = PythonOperator(task_id="reconstruct-narrowband-2", python_callable=null, dag=dag)

(recon_bb, data_alignment1) >> recon_nb1
(recon_bb, data_alignment2) >> recon_nb2


# """
# Calibration tasks
# """
# prefil_cal = PythonOperator(task_id="correct-for-prefilter", bash_command="", dag=dag)

# (prefil_red, recon_nb1, recon_nb2) >> prefil_cal

# align2 = PythonOperator(task_id="realign-channels", bash_command="", dag=dag)

# prefil_cal >> align2

# wlshift = PythonOperator(task_id="correct-for-wavelength-shift", bash_command="", dag=dag)
# align2 >> wlshift


# """
# Science processing tasks
# """
# demod = PythonOperator(task_id="demodulate-pol-states", bash_command="", dag=dag)

# (wlshift, polcal) >> demod

# tmatrix = PythonOperator(task_id="correct-for-tmatrix", bash_command="", dag=dag)

# demod >> tmatrix

# crosstalk = PythonOperator(task_id="correct-for-crosstalk", bash_command="", dag=dag)

# tmatrix >> crosstalk

# combine = PythonOperator(task_id="combine-narrowbands", bash_command="", dag=dag)

# crosstalk >> combine



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# Packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._dkist_init import *  # noqa: F403
# ----------------------------------------------------------------------------

if not _ASTROPY_SETUP_:  # noqa: F405
    # For egg_info test builds to pass, put package imports here.
    from .example_mod import *  # noqa: F401,F403

import logging
import logging.config


logging.basicConfig(level=logging.WARNING)
logging.config.dictConfig(
    {
        "version": 1,
        "loggers": {"astropy.nddata.ccddata": {"level": "WARNING"}},
        "disable_existing_loggers": True,
    }
)



"""
Prefilter frame reduction

Reduce prefilter frames to curve coefficients which approximate the instrumental effects of the
etalon.

See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

import click


@click.command()
def reduce_prefilter(prefilterfiles):
    """
    Function to reduce, correct and process VTF prefilter frames

    Loads specified prefilter frames and determines the parameters of the prefilter curve. This is
    achieved by performing the following operations:

    - Apply dark-correction to frames
    - Average over accumulations
    - Correct for wavelength shifts using the wavelength-shift map calculated during flat-frame
      reduction
    - Average each frame to produce an average spectrum
    - Determine wavelength offset between VTF spectrum and FTS atlas spectrum and adjust for it
    - Determine ratio between VTF and FTS spectra
    - Fit a curve to the ratio of spectra using a double-Gaussian or Lorentzian profile

    \b
    Parameters
    ----------
        List of `pathlib.Path<https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_ objects
        which each point to a FITS file containing a raw VTF prefilter frame.

    \b
    Returns
    -------
    prefilter_params : dict
        Dictionary of parameters describing the profile fitted to the prefilter spectrum.

    """


# If this script is invoked from the command line, run reduced_prefilter()
# Command line arguments are passed automatically by click
if __name__ == "__main__":
    reduce_prefilter()



"""
Polarisation calibration preprocessing

Tools to prepare the polarisation calibration data for processing by DKIST.

"""

import click


@click.command()
def preprocess_polcal(polcal_frames):
    """
    Function to preprocess polarisation calibration data.

    Loads specified polarisation calibration frames, corrects them for darks, and averages over
    accumulations.

    \b
    Parameters
    ----------
    polcal_frames : list
        List of `pathlib.Path<https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_ objects
        which each point to a FITS file containing a raw VTF polarisation calibration frame.

    \b
    Returns
    -------
    reduced_polcal_frames : list
        List of `pathlib.Path<https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_ objects
        which each point to a FITS file containing a preprocessed VTF polarisation calibration frame
        corrected for darks and averaged over accumulations.

    """


# If this script is invoked from the command line, run reduced_prefilter()
# Command line arguments are passed automatically by click
if __name__ == "__main__":
    preprocess_polcal()



import logging
from os.path import join, split
from pathlib import Path
from itertools import groupby
from collections import defaultdict

import numpy as np

import asdf
import ccdproc
from astropy.io import fits

from dkist.asdf_maker import headers_from_filenames as heads
from dkist.asdf_maker import references_from_filenames as refs


def get_modstate(filepath):
    header = fits.getheader(filepath)
    m = header["VTF__031"]
    return f"modstate{int(m)}"


def get_channel(filepath):
    channels = defaultdict(lambda: None)
    channels["BB"] = "broadband"
    channels["NB1"] = "narrowband1"
    channels["NB2"] = "narrowband2"
    header = fits.getheader(filepath)
    if "VTF__002" in list(header.keys()):
        c = header["VTF__002"]
        if c != "default":
            return channels[c]

    for c in ["broadband", "narrowband1", "narrowband2"]:
        if c in filepath.parts:
            return c


def mask_aperture(arr):
    x, y = arr.shape
    x0, y0 = np.array((x, y)) / 2
    xx, yy = np.mgrid[:x, :y]
    r = np.hypot(xx - x0, yy - y0)
    maxr = min(x0, y0) * 0.98
    arr[r > maxr] = np.nan

    return arr


def average_by_wavelength(asdf_file, filekey):
    """
    Function to average VTF flat frames

    Loads specified flat frames, groups them by wavelength position and averages them, producing one
    averaged flat per wavelength step for the narrowband channels and one master flat for the
    broadband channel. Input files should all be for the same line scan and VTF channel. Reduced
    frames are saved to the output directory specified by `data_tree` and references to the files
    are added to the tree for use later in the calibration process.

    Parameters
    ----------
    data_tree : string or `Path<https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_
        Path to an `AsdfFile` defining the calibration data structure, including input and output
        data directories. See :meth:`vtf-pipeline.commands.init_data_tree` for generating an
        appropriate file.

    Returns
    -------
    new_asdf : :class:`asdf.AsdfFile`
        Updated :class:`asdf.AsdfFile` containing the same information as `data_tree` plus file
        references to the averaged flat frames.

    Examples
    --------

    """
    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")

    # TODO This is always going to be something that needs ensuring, so consider making a decorator for it
    if asdf_file["mode"] not in ["broadband", "narrowband"]:
        raise ValueError(
            f'Unrecognised data mode in asdf tree: {asdf_file["mode"]} '
            '- "mode" keyword should be either "broadband" or "narrowband".'
        )

    raw_dir = asdf_file["raw"]["data_dir"]
    logger.debug(f"{raw_dir}")
    outdir = asdf_file["support"]["data_dir"]
    logger.debug(f"{outdir}")

    # This is not especially robust
    prefix = filekey[:-1]
    logger.debug(f"Searching for raw {prefix} files in {raw_dir}")
    fnames = list(Path(raw_dir).rglob(f"{prefix}*.FITS"))
    logger.debug(f"Found the following raw {prefix} files: {[str(f) for f in fnames]}")

    fnames = sorted(fnames, key=get_modstate)
    filegroups = groupby(fnames, key=get_modstate)
    for modstate, files in filegroups:
        logger.debug(f"{modstate} {files}")
        averaged_frames = []
        filelist = list(files)
        modfiles = refs(filelist, np.array(heads(filelist)), len(filelist))
        # TODO rewrite the above as a dict comprehension using the walrus operator (requires py3.8)
        # files_by_modstate = {modstate: refs(flist:=list(files), np.array(heads(flist)), len(flist)) for modstate, files in groupby(sorted(fnames, key=get_modstate), key=get_modstate)}

        fnames = groupby(modfiles, key=lambda x: split(x.fileuri)[-1])
        for out_fname, files in fnames:
            # TODO: consider using ImageFileCollection here instead
            frames = [
                ccdproc.CCDData.read(f.fileuri, format="fits", unit="adu")
                for f in list(files)
            ]
            outfile = join(outdir, modstate, out_fname)
            logger.debug(f"Averaging {filekey} to {outfile}")
            ccdproc.combine(frames, output_file=outfile)
            averaged_frames.append(outfile)
        averaged_frames.sort()
        logger.debug(averaged_frames)

        headers = np.array(heads(averaged_frames))
        data_key = f"reduced averaged {filekey}" + f" {modstate}"
        logger.debug(f"Inserting frames into data tree as '{data_key}'")
        asdf_file["support"][data_key] = refs(
            averaged_frames, headers, len(averaged_frames)
        )

    asdf_file.update()

    return asdf_file


def average_all(asdf_file, filekey):
    """
    Function to average VTF flat frames

    Loads specified flat frames, groups them by wavelength position and averages them, producing one
    averaged flat per wavelength step for the narrowband channels and one master flat for the
    broadband channel. Input files should all be for the same line scan and VTF channel. Reduced
    frames are saved to the output directory specified by `data_tree` and references to the files
    are added to the tree for use later in the calibration process.

    Parameters
    ----------
    data_tree : string or `Path<https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_
        Path to an `AsdfFile` defining the calibration data structure, including input and output
        data directories. See :meth:`vtf-pipeline.commands.init_data_tree` for generating an
        appropriate file.

    Returns
    -------
    new_asdf : :class:`asdf.AsdfFile`
        Updated :class:`asdf.AsdfFile` containing the same information as `data_tree` plus file
        references to the averaged flat frames.

    Examples
    --------

    """
    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")

    # TODO This is always going to be something that needs ensuring, so consider making a decorator for it
    if asdf_file["mode"] not in ["broadband", "narrowband"]:
        raise ValueError(
            f'Unrecognised data mode in asdf tree: {asdf_file["mode"]} '
            '- "mode" keyword should be either "broadband" or "narrowband".'
        )

    raw_dir = asdf_file["raw"]["data_dir"]
    outdir = asdf_file["support"]["data_dir"]
    logger.debug(f"{outdir}")

    frames = [
        ccdproc.CCDData.read(f.fileuri, format="fits", unit="adu")
        for f in asdf_file["raw"][filekey]
    ]
    outfile = join(outdir, f"{filekey[:-1]}.FITS")
    logger.debug(f"Averaging {filekey} to {outfile}")
    ccdproc.combine(frames, output_file=outfile)

    headers = np.array(heads([outfile]))
    data_key = f"reduced averaged {filekey}"
    logger.debug(f"Inserting frames into data tree as '{data_key}'")
    asdf_file["support"][data_key] = refs(outfile, headers, 1)

    asdf_file.update()

    return asdf_file


def filenames_by_fits_keyword(path, keyword, value):
    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")
    allfiles = list(path.rglob("*.FITS"))
    allheaders = heads(allfiles)
    files = []

    logger.debug(f"Checking for keyword {keyword} = {value}")
    for f, h in zip(allfiles, allheaders):
        if keyword in h.keys() and h[keyword] == value:
            files.append(f)
    logger.debug(f"Found {len(files)} files")

    return files


def add_files_by_keywords(asdf_file, dkey, kw_dict, raw=False):
    """
    """

    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")
    logger.debug(f"Finding files with keywords {kw_dict}")
    keybase = "raw" if raw else "support"
    data_dir = asdf_file[keybase]["data_dir"]
    logger.debug(f"Searching {data_dir}")
    if isinstance(data_dir, str):
        data_dir = Path(data_dir)
    allfiles = None
    for keyword in kw_dict.keys():
        files = filenames_by_fits_keyword(data_dir, keyword, kw_dict[keyword])
        # filter by overlap with previous files
        if allfiles:
            allfiles = list(set(files) & set(allfiles))
        else:
            allfiles = files
    allfiles = sorted(allfiles)

    logger.debug(f"Selected {len(allfiles)} files")
    asdf_file[keybase][dkey] = refs(allfiles, np.array(heads(allfiles)), len(allfiles))
    try:
        asdf_file.update()
    except ValueError:
        f = asdf_file.uri
        asdf_file.write_to(f)
        asdf_file.close()
        asdf_file = asdf.open(f, mode="rw")
    return asdf_file



import re
import logging
from os.path import join

import matplotlib.pyplot as plt

import asdf
from astropy.io import fits

from vtfcal.calibration.demodulate import demod
from vtfcal.test_constants import TEST_WL_IDX
from vtfcal.utils import plotframes


def demodulate(data_tree):
    """
    """

    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    asdf_file = asdf.open(data_tree, mode="rw")
    outdir = asdf_file["support"]["data_dir"]

    asdf_file = demod(asdf_file)
    asdf_file.update()

    titles = [r"$I_0$", r"$I_1$", r"$I_2$", r"$I_3$", r"$I$", r"$Q$", r"$U$", r"$V$"]
    keys = [f"calibrated aligned data modstate{m}" for m in range(4)] + [f"calibrated demodulated data modstate{m}" for m in range(4)]
    plotframes(asdf_file, list(zip(keys, titles)), "05-demodulated-data")



"""
Dark frame reduction

Average collected dark frames as appropriate to reduce to the smallest number required to calibrate
the science data.

See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

from os.path import join

import asdf

from vtfcal.utils import add_files_by_keywords


def init_data_tree(indir):
    """
    Initiate a set of data trees for describing the calibration data structure.

    Creates and saves an :class:`~asdf.AsdfFile` for each VTF camera and initialises them with
    `inidir` and `outdir` as input and output directories, respectively. Also defines for each file
    whether the data are broadband or narrowband images.

    \b
    Parameters
    ----------
    indir : str or :class:`pathlib.Path`
        Directory in which raw input is stored.

    \b
    outdir : str or :class:`pathlib.Path`
        Directory to which intermediate and final output files will be saved.
    """

    channels = [("bb", "broadband"), ("nb1", "narrowband1"), ("nb2", "narrowband2")]

    for short, long_ in channels:
        tree = {
            # Would use pathlib for these but asdf can't represent it
            "raw": {"data_dir": join(indir, "raw", long_)},
            "support": {"data_dir": join(indir, "support", long_)},
            "processed": {"data_dir": join(indir, "processed", long_)},
            "polarization": {"data_dir": join(indir, "raw", "polarization")},
            "plots": join(indir, "plots", long_),
            "mode": "broadband" if short == "bb" else "narrowband",
            "speckle": {"data_dir": join(indir, "raw", "speckle")},
        }

        f = join(indir, f"{short}_data_tree")
        asdf_file = asdf.AsdfFile(tree, uri=f)
        asdf_file = add_files_by_keywords(asdf_file, "darks", {"DKIST003": "dark"}, raw=True)
        # Header key indicating channel isn't in the format I'm using for tracking files
        chankey = short.upper().replace("1", "+").replace("2", "-")
        for dkey, head_kw in zip(["flats", "data"], ["gain", "observe"]):
            for m in range(4):
                asdf_file = add_files_by_keywords(
                    asdf_file,
                    f"{dkey} modstate{m}",
                    {"VTF__002": chankey, "VTF__031": m, "DKIST003": head_kw},
                    raw=True,
                )



# import os

# If this package has tests data in the tests/data directory, add them to
# the paths here, see commented example
paths = ['coveragerc',
#         os.path.join('data', '*fits')
         ]

def get_package_data():
    return {_ASTROPY_PACKAGE_NAME_ + ".tests": paths}



"""
Frame alignment

Functions to align frames based on coalignment parameters calculated by target reduction

See Also
--------

Notes
-----

References
----------

Examples
--------

"""

import re
import logging
from os.path import join

import matplotlib.pyplot as plt
import numpy as np
from skimage.transform import AffineTransform, warp

import asdf
from astropy.io import fits

from dkist.asdf_maker import headers_from_filenames as heads
from dkist.asdf_maker import references_from_filenames as refs
from vtfcal.test_constants import TEST_PIXEL, TEST_WL_IDX
from vtfcal.utils import plotframes


def align_frames(data_tree):
    """
    """
    logger = logging.getLogger(__name__)

    asdf_file = asdf.open(data_tree, mode="rw")

    tform = AffineTransform(matrix=asdf_file["coalignment-matrix"])

    for modstate in ["modstate0", "modstate1", "modstate2", "modstate3"]:
        aligned_frames = []
        datafiles = asdf_file["support"]["corrected flat-corrected data " + modstate]
        for f in datafiles:
            logger.debug(f.fileuri)
            hdu = fits.open(f.fileuri)[0]
            img = hdu.data
            new_img = warp(img, tform.inverse, output_shape=img.shape, cval=img.mean())
            hdu.data = new_img
            # TODO also update the rotation in the header because that will obviously be important
            new_fname = f.fileuri.replace("flat_corrected", "aligned")
            fits.writeto(new_fname, hdu.data, header=hdu.header)
            aligned_frames.append(new_fname)

        headers = np.array(heads(aligned_frames))
        asdf_file["support"]["calibrated aligned data " + modstate] = refs(
            aligned_frames, headers, len(aligned_frames)
        )

        plotframes(
            asdf_file,
            [
                (f"corrected flat-corrected data {modstate}", "Flat-corrected frame"),
                (
                    f"calibrated aligned data {modstate}",
                    "Derotated flat-corrected frame",
                ),
            ],
            f"04a-aligned-data-{modstate}",
        )

    asdf_file.update()



# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure.

from astropy.version import version as astropy_version

if astropy_version < "3.0":
    # With older versions of Astropy, we actually need to import the pytest
    # plugins themselves in order to make them discoverable by pytest.
    from astropy.tests.pytest_plugins import *
else:
    # As of Astropy 3.0, the pytest plugins provided by Astropy are
    # automatically made available when Astropy is installed. This means it's
    # not necessary to import them here, but we still need to import global
    # variables that are used for configuration.
    from astropy.tests.plugins.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS

from astropy.tests.helper import enable_deprecations_as_exceptions

## Uncomment the following line to treat all DeprecationWarnings as
## exceptions. For Astropy v2.0 or later, there are 2 additional keywords,
## as follow (although default should work for most cases).
## To ignore some packages that produce deprecation warnings on import
## (in addition to 'compiler', 'scipy', 'pygments', 'ipykernel', and
## 'setuptools'), add:
##     modules_to_ignore_on_import=['module_1', 'module_2']
## To ignore some specific deprecation warning messages for Python version
## MAJOR.MINOR or later, add:
##     warnings_to_ignore_by_pyver={(MAJOR, MINOR): ['Message to ignore']}
# enable_deprecations_as_exceptions()

## Uncomment and customize the following lines to add/remove entries from
## the list of packages for which version numbers are displayed when running
## the tests. Making it pass for KeyError is essential in some cases when
## the package uses other astropy affiliated packages.
# try:
#     PYTEST_HEADER_MODULES['Astropy'] = 'astropy'
#     PYTEST_HEADER_MODULES['scikit-image'] = 'skimage'
#     del PYTEST_HEADER_MODULES['h5py']
# except (NameError, KeyError):  # NameError is needed to support Astropy < 1.0
#     pass

## Uncomment the following lines to display the version number of the
## package rather than the version number of Astropy in the top line when
## running the tests.
# import os
#
## This is to figure out the package version, rather than
## using Astropy's
# try:
#     from .version import version
# except ImportError:
#     version = 'dev'
#
# try:
#     packagename = os.path.basename(os.path.dirname(__file__))
#     TESTED_VERSIONS[packagename] = version
# except NameError:   # Needed to support Astropy <= 1.0.0
#     pass



import os
import re
import shutil
import logging
import tempfile
import configparser as cfp
from pathlib import Path

import numpy as np
from skimage.transform import AffineTransform, warp

from astropy.io import fits

from vtfcal import utils


def process(data_dir=utils.DATA, raw_data_dir=utils.DATA, dataset_id=None):
    """"""
    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")
    # This bit just to convert raw speckle calibration burst into fits
    # Probably no longer necessary
    specburst = raw_data_dir / "speckle"
    specburst_files = specburst.rglob("*.raw")
    for fname in specburst_files:
        img = np.fromfile(fname, dtype="int16").reshape((1024, 1024))
        newf = raw_data_dir / "broadband" / f"speckle_{fname.stem}.FITS"
        fits.writeto(str(newf), img, overwrite=True)

    cfg = cfp.ConfigParser()
    cfg.read("vtf-calibration-config.cfg")
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)
    dataset_dir = (
        data_dir / dataset_id if dataset_id else Path(tempfile.mkdtemp(dir=data_dir))
    )
    if dataset_id:
        s = dataset_dir / "support"
        oldfiles = s.rglob("*.FITS")
        [os.remove(f) for f in oldfiles]

    allfiles = raw_data_dir.rglob("*.FITS")

    chankeys = {"broadband": "BB", "narrowband1": "NB+", "narrowband2": "NB-"}
    for channel in ["broadband", "narrowband1", "narrowband2", "polarization"]:
        subdirs = ["raw", "processed", "support", "polarization", "plots"]
        for subdir in subdirs:
            s = dataset_dir / subdir / channel
            if not os.path.exists(s):
                os.makedirs(s)
            if subdir != "plots":
                for m in range(4):
                    s2 = s / f"modstate{m}"
                    if not os.path.exists(s2):
                        os.makedirs(s2)

    for rawfile in allfiles:
        rawpath = rawfile.relative_to(raw_data_dir).parent
        newfile = str(dataset_dir / "raw" / rawpath / f"{rawfile.stem}.FITS")
        channel = utils.get_channel(rawfile)

        if (rawfile.name == "target.FITS") or (channel is None):
            shutil.copy2(rawfile, newfile)
            print(f"copying {rawfile} => {newfile} (channel = {channel})")
        else:
            # TODO Take out the hard-coded image sizes here and use TEST_IMAGE_SIZE instead
            raw_image = fits.open(rawfile)[0]
            img = raw_image.data
            rot = cfg.getfloat("camera-rotations", channel)
            scale = 512 / np.array(img.shape)
            c = np.cos(np.deg2rad(rot))
            s = np.sin(np.deg2rad(rot))
            rmatrix = np.array([[c, -s, 0], [s, c, 0], [0, 0, 1]])
            array_centre = np.array([512, 512]) / 2.0
            displacement = np.dot(rmatrix[:2, :2], array_centre) - array_centre
            rmatrix[:2, 2] = -displacement
            rmatrix[:2, :2] *= scale
            tform = AffineTransform(matrix=rmatrix)
            new_image = warp(img, tform.inverse, output_shape=(512, 512), cval=img.mean())
            new_image = utils.mask_aperture(new_image)

            # Raw simulation data don't distinguish between channels at the moment
            # TODO double check that they do distinguish between modstates
            new_header = raw_image.header
            new_header["VTF__002"] = chankeys[channel]

            fits.writeto(
                newfile, new_image, header=new_header, overwrite=True
            )

    return dataset_dir



"""
Flat frame reduction

Average collected flat frames as appropriate to reduce to the smallest number required to calibrate
the science data.

See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

import logging
from os.path import join
from pathlib import Path

import dask.array as da
import matplotlib.pyplot as plt
import numpy as np
from scipy.ndimage.interpolation import shift

from astropy.io import fits
from astropy.io.fits.hdu.base import BITPIX2DTYPE
from astropy.modeling import fitting, models

from dkist.asdf_maker import headers_from_filenames as heads
from dkist.asdf_maker import references_from_filenames as refs
from dkist.io import DaskFITSArrayContainer as DFAC
from dkist.io.fits import AstropyFITSLoader as Loader
from vtfcal.test_constants import TEST_PIXEL, TEST_WL_IDX
from vtfcal.utils import average_by_wavelength

# from dask.distributed import get_client  # Client


def get_dtype_and_size(f):
    with f as fi:
        head = fits.getheader(fi, hdu=1)
    naxes = head["NAXIS"]
    dtype = BITPIX2DTYPE[head["BITPIX"]]
    shape = [head[f"NAXIS{n}"] for n in range(naxes, 0, -1)]
    return dtype, shape


class DelayedFits:
    def __init__(self, file, shape, dtype):
        self.shape = shape
        self.dtype = dtype
        self.file = file
        self.hdu = 0

    def __getitem__(self, item):
        with self.file as f:
            with self.file as f:
                with fits.open(f) as hdul:
                    hdul.verify("fix")
                    return hdul[self.hdu].data[item]


def average_flats(asdf_file):
    return average_by_wavelength(asdf_file, "flats")


def calculate_wl_shift(asdf_file, modstate, fourier=True):
    """
    Calculate the wavelength shift of each pixel from the image mean profile

    Extract a wavelength profile and determine its minimum, both for each individual pixel position
    and for the average across all pixels in each image. The wavelength shift for each pixel is
    calculated as the difference between these two values.

    Parameters
    ----------
    asdf_file : string or `Path<https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_
        Path to an `AsdfFile` defining the calibration data structure, including input and output
        data directories and file references to averaged, dark-corrected flat frames. See
        :meth:`vtf-pipeline.commands.init_data_tree` for generating an appropriate file.

    Returns
    -------
    new_asdf : :class:`asdf.AsdfFile`
        Updated :class:`asdf.AsdfFile` containing the same information as `data_tree` plus a file
        reference to a map of the wavelength shift value calculated for each pixel.
    """
    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")

    # client = get_client()

    outdir = Path(asdf_file["support"]["data_dir"])

    ## REFACTOR: file loading
    flatfiles = asdf_file["support"]["corrected dark-corrected flats " + modstate]
    arr = DFAC(flatfiles, loader=Loader).array
    arr = arr.rechunk((-1, 8, 8))
    mean_profile = da.nanmean(arr, axis=(1, 2)).compute()
    profile = arr[:, TEST_PIXEL[0], TEST_PIXEL[1]].compute()

    if fourier:
        flats = np.array(arr)
        flats[np.isnan(flats)] = 0
        l = flats.shape[0]
        flats_fft = np.fft.fftn(flats, axes=[0])
        lp = -np.arctan(flats_fft[1].imag / flats_fft[1].real) / (2 * np.pi)
        wl_shift_map = -(lp * l)
    else:
        # cropped_profile = profile[profile.argmin()-5:profile.argmin()+6]
        model = models.Polynomial1D(degree=2)
        fitter = fitting.LinearLSQFitter()
        # xp = range(profile.argmin()-5, profile.argmin()+6)
        # fit_profile = fitter(model, xp, cropped_profile)
        fit_profile = fitter(model, range(len(profile)), profile)
        c, b, a = fit_profile.parameters
        minx = -b / (2 * a)
        miny = c - ((b ** 2) / (4 * a))
        logger.debug(f"\n{fit_profile}\n{c, b, a}\n{minx, miny}")

        minmap = da.from_array(np.zeros((arr.shape[1], arr.shape[2])), chunks=(8, 8))
        minmap = da.map_blocks(
            _calcshift, minmap, arr, dtype=minmap.dtype, chunks=(1, 8, 8)
        )
        minmap = minmap.reshape((arr.shape[1], arr.shape[2]))

        wl_shift_map = da.nanmean(minmap) - minmap

    if fourier:
        final_map = wl_shift_map
    else:
        final_map = wl_shift_map.compute()
    fname = outdir / "wl_shift_map.FITS"
    fits.writeto(fname, final_map, overwrite=True)
    asdf_file["support"]["calibration wl-shift-map " + modstate] = refs(
        [fname], np.array(heads([fname])), 1
    )
    asdf_file.update()

    return asdf_file, arr, mean_profile


# TODO Consider moving this function and _shift_pixels into .utils somewhere
def _calcshift(wl_shift_map, flatfiles):
    """
    Wavelength-shift calculation

    Fits a second-order polynomial to the wavelength profile of each pixel in `flatfiles`, and
    stores the minima of the polynomials in `wl_shift_map`. Only uses a small range around the
    minimum value of the wavelength profile, since including the wings can produce inaccurate fits.

    Parameters
    ----------
    wl_shift_map : :class:`dask.array.Array`
        Empty array with the same shape as flat frames.

    flatfiles : :class:`dask.array.Array`
        3D array containing flat frame values for each wavelength step

    Returns
    -------
    wl_shift_map : :class:`dask.array.Array`
        Array of calculated profile minima for each pixel
    """
    logger = logging.getLogger(__name__)

    l, xx, yy = flatfiles.shape

    ## OPTIMISE: remove loops if possible
    fitter = fitting.LinearLSQFitter()
    for x in range(xx):
        for y in range(yy):
            profile = flatfiles[:, x, y]
            xc = profile.argmin()
            model = models.Polynomial1D(degree=2)
            try:
                ## Hard-coded indices assume wl range of scan will be small
                ## Double check this and see if header info can be used instead
                fit_profile = fitter(
                    model, range(xc - 5, xc + 6), profile[xc - 5 : xc + 6]
                )
                c, b, a = fit_profile.parameters
                minx = -b / (2 * a)
            ## OPTIMISE: find a way to exclude these pixels without try/except
            except Exception as e:
                logger.debug(
                    (
                        x,
                        y,
                        xc,
                        len(range(xc - 5, xc + 6)),
                        len(profile[xc - 5 : xc + 6]),
                        len(range(xc - 5, xc + 6)) == len(profile[xc - 5 : xc + 6]),
                    )
                )
                logger.debug(f"Failed to fit pixel {x}, {y} of chunk with error: {e}")
                minx = np.nan
            wl_shift_map[x, y] = minx

    return wl_shift_map


def _shift_pixels(flatfiles, wl_shift_map):
    """
    Shift wavelength profiles

    Interpolates the wavelength profile of each pixel along the wavelength axis to align the minimum
    of the profile with the minimum of the average profile.

    Parameters
    ----------
    flatfiles : :class:`dask.array.Array`
        3D array containing flat frame values for each wavelength step

    wl_shift_map : :class:`dask.array.Array`
        Wavelength shifts calculated by :meth:`_calcshift`
    """
    logger = logging.getLogger(__name__)

    xx, yy = wl_shift_map.shape

    ## OPTIMISE: remove loop
    for x in range(xx):
        for y in range(yy):
            if np.isnan(wl_shift_map[x, y]):
                logger.debug(f"No valid shift value for pixel {x}, {y}, skipping.")
                flatfiles[:, x, y] = np.nan
            else:
                flatfiles[:, x, y] = shift(
                    flatfiles[:, x, y], wl_shift_map[x, y], mode="nearest"
                )
            ## This wants to be something more accurate than nearest once i've optimised a bit

    return flatfiles


def correct_wl_shift(
    flatfiles, wl_shift_map, mean_profile, outdir, modstate, fourier=True
):
    """
    Correct flat-field images by adjusting for wavelength shift

    Interpolate the given flat-field images such that the value in each pixel position is adjusted
    by an amount specified by the wavelength shift map. This aligns the centre of the wavelength
    profile of each pixel.

    Parameters
    ----------
    flatfiles : array-like
        Dask array of per-wavelength-position
        dark-corrected flats. For the broadband channel this dictionary should contain a single
        frame.

    wl_shift_map : array of floats
        Array containing the wavelength shift value calculated for each pixel.

    mean_profile : array
        Wavelength profile averaged over all pixels. Used only for plotting.

    outdir : string
        Output directory in which to save plots

    Returns
    -------
    corrected_flats : array-like
        Dask array of per-wavelength-position flats corrected for wavelength shift.

    Raises
    ------

    See Also
    --------
    calculate_wl_shift

    Notes
    -----

    References
    ----------
    Optional, use if references are cited in Notes

    Examples
    --------

    """
    if fourier:
        wl_shift_map[np.isnan(wl_shift_map)] = 0
        flats = np.array(flatfiles)
        nanidx = np.isnan(flats)
        flats[nanidx] = 0
        flats_fft = np.fft.fftn(flats, axes=[0])
        l = flats_fft.shape[0]
        midpoint = l // 2
        k = np.roll(-(np.arange(l) - midpoint), midpoint)
        d = 2 * np.pi * (k / l) * wl_shift_map.reshape(*wl_shift_map.shape, 1)
        d = np.rollaxis(d, 2)
        d1 = np.cos(d) + (np.sin(d) * 1j)
        corrected_flats = np.fft.ifftn(flats_fft * d1, axes=[0]).real
        corrected_flats[nanidx] = np.nan
        corrected_flats = da.from_array(corrected_flats)
    else:
        wl_shift_map = da.from_array(wl_shift_map, chunks=(8, 8))
        corrected_flats = da.map_blocks(
            _shift_pixels, flatfiles, wl_shift_map, dtype=flatfiles.dtype
        )

    profile = flatfiles[:, TEST_PIXEL[0], TEST_PIXEL[1]]
    shifted_profile = corrected_flats[:, TEST_PIXEL[0], TEST_PIXEL[1]]

    ## REFACTOR: plotting helpers
    fig, ax = plt.subplots(figsize=(10, 7))
    plt.plot(profile, label=f"Unshifted profile for pixel {TEST_PIXEL}")
    plt.axhline(profile.min(), linestyle=":")
    if not fourier:
        plt.plot(mean_profile, linestyle="--", color="red", label="Mean profile")
        plt.plot(mean_profile.argmin(), mean_profile.min(), "x", color="red")
        target = mean_profile.argmin()
    else:
        target = midpoint
    plt.axvline(target, linestyle="--", color="red")
    plt.plot(
        shifted_profile,
        linestyle=":",
        color="green",
        label=f"Shifted profile for pixel {TEST_PIXEL}",
    )
    plt.axvline(shifted_profile.argmin(), linestyle=":", color="green")
    plt.xlabel(r"$\lambda$ index")
    plt.ylabel(r"$I$")
    plt.legend()
    testshift = wl_shift_map[TEST_PIXEL[0], TEST_PIXEL[1]]
    if not fourier:
        testshift = testshift.compute()
    plt.title(
        f"Profile minimum shifted from {profile.argmin().compute()} to {shifted_profile.argmin().compute()}\n(Target = {target}, calculated shift = {testshift})"
    )
    plt.savefig(join(outdir, f"01e-wl-shifted-profile-{modstate}"), bbox_tight="tight")
    plt.close()

    return corrected_flats



"""
Short summary

Extended summary

See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

import logging
from pathlib import Path

import dask.array as da
import numpy as np

from astropy.io import fits

from dkist.asdf_maker import headers_from_filenames as heads
from dkist.asdf_maker import references_from_filenames as refs
from dkist.io import DaskFITSArrayContainer as DFAC
from dkist.io.fits import AstropyFITSLoader as Loader


def demod(asdf_file):
    """
    Apply demodulation matrix to reconstructed data frames.
    """

    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")

    outdir = asdf_file["support"]["data_dir"]

    modstates = ["modstate0", "modstate1", "modstate2", "modstate3"]
    data_by_modstate = {
        m: asdf_file["support"]["calibrated aligned data " + m] for m in modstates
    }
    demod_by_modstate = {m: [] for m in modstates}

    # for each wavelength
    for l in range(18):
        # load frame for each mod state (x, y, 4)
        allstates = DFAC(
            [data_by_modstate[m][l] for m in modstates], loader=Loader
        ).array
        ## Moving the axis in this way _shouldn't_ make anything wrong, but make sure of that at some point
        allstates = da.moveaxis(allstates, 0, -1)
        logger.debug(
            f"{allstates.shape}, {type(allstates)}, {allstates.mean().compute()}"
        )
        logger.debug(asdf_file["polcal-demod-matrices"])
        # This isn't right at the moment because Athur's code makes a 4x8 matrix
        # demod_mat = fits.open(asdf_file["polcal-demod-matrices"])[1]
        # TODO Double check that this array is a) the right way round and b) in the right order
        rt34 = np.sqrt(3) / 4
        demod_mat = np.array(
            [
                [1 / 4, 1 / 4, 1 / 4, 1 / 4],
                [rt34, rt34, -rt34, -rt34],
                [rt34, -rt34, -rt34, rt34],
                [rt34, -rt34, rt34, -rt34],
            ]
        )
        logger.debug(f"{demod_mat.shape}")
        newdata = np.matmul(allstates, demod_mat)  # .data)
        logger.debug(f"{newdata.shape}, {type(newdata)}")

        for m, mod in enumerate(modstates):
            # get/define output name
            outfile = data_by_modstate[mod][l].fileuri.replace("aligned", "demodulated")
            logger.debug(f"{outfile}")
            # save file
            # This will want a defined header
            fits.writeto(outfile, np.array(newdata[..., m]), overwrite=True)
            # put filname in list
            demod_by_modstate[mod].append(outfile)

    for mod, files in demod_by_modstate.items():
        asdf_file["support"]["calibrated demodulated data " + mod] = refs(
            files, np.array(heads(files)), len(files)
        )

    asdf_file.update()

    return asdf_file



import warnings

from vtfcal.utils import RAW_DATA
from vtfcal.utils.preprocess_data import process

from vtfcal.pipeline import (align_frames, correct_darks, correct_flats, demodulate,
                             init_data_tree, reconstruct, reconstruct_broadband, reduce_darks,
                             reduce_flats, reduce_polcal, reduce_targets)

warnings.filterwarnings("ignore", category=UserWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)

dsid = "00000000"

# Perform initial pre-processing of simulated data
# Provides the proper folder structure, offsets frame alignment, and scales for reasonable run time
print("Preprocessing simulation data...", end="")
process(dataset_id=dsid)
print("Done")

# Initiate data tree which tracks filenames and types of data frames
print("Initiating data tree...", end="")
init_data_tree(DATA)
print("Done")
bb_data_tree = dataset_dir / "bb_data_tree"
nb1_data_tree = dataset_dir / "nb1_data_tree"
nb2_data_tree = dataset_dir / "nb2_data_tree"

# Reduce darks
print("Reducing dark frames...", end="")
reduce_darks(bb_data_tree)
reduce_darks(nb1_data_tree)
reduce_darks(nb2_data_tree)
print("Done")

# Reduce flats
print("Reducing flat frames...", end="")
reduce_flats(bb_data_tree)
reduce_flats(nb1_data_tree, correction=True, fourier=True)
reduce_flats(nb2_data_tree, correction=True, fourier=True)
print("Done")

# Reduce targets
print("Reducing target frames...", end="")
reduce_targets([bb_data_tree, nb1_data_tree, nb2_data_tree])
print("Done")

# Reduce prefilter not implemented

# Polarimetric calibration
print("Reducing polarimetric calibration frames...", end="")
reduce_polcal(nb1_data_tree)
reduce_polcal(nb2_data_tree)
print("Done")

# Dark correction
print("Dark-correcting science frames...", end="")
correct_darks(bb_data_tree)
correct_darks(nb1_data_tree)
correct_darks(nb2_data_tree)
print("Done")

# Flat correction
print("Flat-correcting science frames...", end="")
correct_flats(bb_data_tree)
correct_flats(nb1_data_tree)
correct_flats(nb2_data_tree)
print("Done")

# Data alignment
print("Aligning narrowband frames to broadband...", end="")
align_frames(nb1_data_tree)
align_frames(nb2_data_tree)
print("Done")

# Broadband reconstruction
# reconstruct_broadband(bb_data_tree)

# Narrowband reconstruction dependent on broadband reconstruction
reconstruct(nb1_data_tree, bb_tree=bb_data_tree)
reconstruct(nb2_data_tree, bb_tree=bb_data_tree)

# Demodulate reconstructed images
print("Demodulating narrowband frames...", end="")
demodulate(nb1_data_tree)
demodulate(nb2_data_tree)
print("Done")



"""
Target frame reduction

Functions to average, dark-correct, flat-correct and align target frames.

See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

import glob
import logging
from os.path import join
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import skimage.feature as skf
from skimage.feature import BRIEF

import asdf

from dkist.asdf_maker import headers_from_filenames as heads
from dkist.asdf_maker import references_from_filenames as refs
from vtfcal.reduction import targets
from vtfcal.utils import correct_darks


def reduce_targets(data_trees):
    """
    Reduce and correct target frames

    Loads target frames from the input directory specified by `data_tree`, then averages them,
    corrects for dark and flat frames, and determines the parameters required to coalign the images.

    Parameters
    ----------
    data_trees : list of strings or :class:`pathlib.Path`s
        Paths to :class:`~asdf.AsdfFile`s defining the calibration data structure, including input
        and output data directories, for each telescope channel. See :meth:`commands.init_data_tree`
        for generating appropriate files.
    """
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    extractor = BRIEF()
    ## Plotting helpers
    fig, ax = plt.subplots(2, 3, figsize=(16, 10))
    ax = ax.flatten()
    for i, fname in enumerate(data_trees):
        for m in range(4):
            modstate = f"modstate{m}"
            channel_tree = asdf.open(fname, mode="rw")
            data_dir = Path(channel_tree["raw"]["data_dir"]) / modstate
            targetfiles = list(data_dir.rglob("target.FITS"))
            headers = np.array(heads(targetfiles))
            channel_tree["raw"]["target {modstate}"] = refs(targetfiles, headers, len(targetfiles))
            channel_tree.update()
            target, target_fname = targets.average_targets(
                targetfiles, Path(channel_tree["support"]["data_dir"]) / modstate
            )

            # existing_reduced = channel_tree.tree.get("reduced", {}).get("averaged", {})
            header = np.array(heads([target_fname]))
            channel_tree["support"][f"reduced averaged target {modstate}"] = refs(target_fname, header, 1)
            channel_tree.update()
            channel_tree = correct_darks(channel_tree, "support", f"reduced averaged target {modstate}")

            # # This isn't plotting dark-corrected targets but that's probably fine for now I guess
            img = target.data
            plt.sca(ax[i])
            plt.imshow(img)
            if channel_tree["mode"] == "broadband":
                points, img_refs = targets.define_alignment_refs(img, extractor)
                ax[i].scatter(
                    points.T[1], points.T[0], marker="x", cmap="Reds", c=range(len(points.T[0]))
                )
            elif channel_tree["mode"] == "narrowband":
                newim, points, matches, matrix = targets.align_to_reference(img, extractor, img_refs)
                ax[i].scatter(
                    points.T[1], points.T[0], marker="x", cmap="Reds", c=range(len(points.T[0]))
                )
                skf.plot_matches(ax[3], img_refs["image"], img, img_refs["points"], points, matches)

                channel_tree["coalignment-matrix"] = matrix
                channel_tree.write_to(channel_tree.uri)
                resid = img_refs["image"] - newim
                logger.debug(f"{resid.max()}")
                ax[i + 3].imshow(resid)
            else:
                raise ValueError(
                    "Unrecognised data mode." "Should be either 'broadband' or 'narrowband'."
                )

    plt.savefig(Path(channel_tree["plots"]).parent / "target-alignment")



import logging
import subprocess as subp

import numpy as np

from astropy.io import fits


def run_speckle(
    width=2048,
    height=2048,
    tile_size=128,
    burst_directory="/home/dkistdev/data",
    burst_prefix="bbso",
    burst_suffix=".pnm",
    burst_size=80,
    fixed_width=0,
    gpus=0,
    master_gpu=0,
    L1=0.949999988,
    L2=0.0900000036,
    iterations=10,
    out_prefix="byproducts",
    fried_parameter=0.100000001,
    wavelength=656.281982,
    pixel_scale=0.0170000009,
    diameter=4,
    dark=None,
    gain=None,
    npsd=None,
    init_with_mean=False,
    floats=False,
):

    logger = logging.getLogger(__name__)
    logger.setLevel("DEBUG")

    # Define bispectrum indices?
    # call spectral ratio library?
    # construct speckle class
    # read darks
    # read flats
    # read 'burst' image files
    # correct for darks and flats
    # sum images?
    # call reconstruction

    if not isinstance(gpus, int):
        gpus = {gpus}

    speckle_command = [
        "./testSpeckle",
        "-w",
        f"{width}",
        "-h",
        f"{height}",
        "-t",
        f"{tile_size}",
        "-d",
        f"{burst_directory}",
        "-p",
        f"{burst_prefix}",
        "-s",
        f"{burst_suffix}",
        "-n",
        f"{burst_size}",
        f"--fixed-width={fixed_width}",
        f"--gpus={gpus}",
        f"--master-gpu={master_gpu}",
        f"--L1={L1}",
        f"--L2={L2}",
        f"--iterations={iterations}",
        f"--fried-parameter={fried_parameter}",
        f"--wavelength={wavelength}",
        f"--pixel-scale={pixel_scale}",
        f"--diameter={diameter}",
    ]
    if dark:
        speckle_command.append(f"--dark={dark}")
    if gain:
        speckle_command.append(f"--gain={gain}")
    if npsd:
        speckle_command.append(f"--npsd={npsd}")
    if init_with_mean:
        speckle_command.append(f"--init-with-mean={init_with_mean}")
    if floats:
        speckle_command.append(f"--floats={floats}")

    subp.run(speckle_command)

    recon = np.fromfile("byproducts-reconstruction.raw", dtype="float32").reshape(
        (1024, 1024)
    )

    return recon


def calc_fried_param(
    dmActs2KlModes="./aotelem/dmActs2KlModes-default-f32.data",
    reconMatrixPrefix="./aotelem/RECON_MAT-",
    reconMatrixSuffix="Modes.data",
    klVariances="./aotelem/klVariances-default-f64.data",
    actCommands="./aotelem/actCommands.data",
    actErrors="./aotelem/actErrors.data",
    shifts="./aotelem/shifts.data",
    mode="1049Modes",
):

    logger = logging.getLogger(__name__)
    logger.setLevel("DEBUG")

    aotelem_command = [
        "./testAOTelemetry",
        f"--dmActs2KlModes={dmActs2KlModes}",
        f"--reconMatrixPrefix={reconMatrixPrefix}",
        f"--reconMatrixSuffix={reconMatrixSuffix}",
        f"--klVariances={klVariances}",
        f"--actCommands={actCommands}",
        f"--actErrors={actErrors}",
        f"--shifts={shifts}",
        f"--mode={mode}",
    ]

    out = subp.run(aotelem_command, capture_output=True, check=True, text=True)

    if "FriedParameter" in out.stdout:
        # This really wants to be somewhat more robust
        result = out.stdout.split("\n")[-2]
        r0 = result.split(" ")[-1]
    else:
        r0 = None

    return r0



"""
Science data preparation

Functions to perform dark corrections on raw science data as part of the preparation for more
rigourous image reconstruction using those data.

See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

import logging

import numpy as np

import asdf
import ccdproc

from dkist.asdf_maker import headers_from_filenames as heads
from dkist.asdf_maker import references_from_filenames as refs
from dkist.io import DaskFITSArrayContainer as DFAC
from dkist.io.fits import AstropyFITSLoader as Loader
from vtfcal.correction.flat_correction import flat_correct
from vtfcal.test_constants import TEST_PIXEL, TEST_WL_IDX
from vtfcal.utils import plotframes, plotprofile


def correct_flats(data_tree):
    """
    Apply flat correction to data frames.

    Loads dark-corrected data frames specified by `data_tree` and corrects them for flat-field
    effects using the reduced flats calculated using :meth:`commands.reduce_flats`. Corrected frames
    are saved to the output directory specified by `data_tree` and references to the files are added
    to the tree for use later in the calibration process.

    \b
    Parameters
    ----------
    data_tree : string or :class:`pathlib.Path`
        Path to an :class:`~asdf.AsdfFile` defining the calibration data structure, including input
        and output data directories, and file references to reduced flats. See
        :meth:`commands.init_data_tree` for generating an appropriate file.

    Examples
    --------

    """
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    asdf_file = asdf.open(data_tree, mode="rw")
    for modstate in ["modstate0", "modstate1", "modstate2", "modstate3"]:
        datafiles = asdf_file["support"]["corrected dark-corrected data " + modstate]
        outdir = asdf_file["support"]["data_dir"]

        if asdf_file["mode"] == "broadband":
            allflats = asdf_file["support"][
                "corrected dark-corrected flats " + modstate
            ]
            flats = [
                ccdproc.CCDData.read(f.fileuri, format="fits", unit="adu")
                for f in allflats
            ]
        else:
            ## REFACTOR: rewrite whatever saves the stack to do it as fits files and as part of the asdf
            flat_data = DFAC(
                asdf_file["support"][f"corrected normalised flats {modstate}"],
                loader=Loader,
            ).array
            flats = [
                ccdproc.CCDData(frame.compute(), unit="adu") for frame in flat_data
            ]
        corrected_frames = flat_correct(datafiles, flats)
        headers = np.array(heads(corrected_frames))
        asdf_file["support"]["corrected flat-corrected data " + modstate] = refs(
            corrected_frames, headers, len(corrected_frames)
        )

        plotframes(
            asdf_file,
            [
                (
                    f"corrected flat-corrected data {modstate}",
                    "Flat-corrected data frame",
                )
            ],
            "02b-flat-corrected-data",
        )

        if asdf_file["mode"] != "broadband":
            plotprofile(
                asdf_file,
                [
                    (f"corrected dark-corrected data {modstate}", "Dark-corrected data"),
                    (f"corrected normalised flats {modstate}", "Normalised flats"),
                    (f"corrected flat-corrected data {modstate}", "Flat-corrected data",),
                ],
                f"02c-flat-correction-comparison-{modstate}",
                linestyle=[":", "--", None],
                color=["black", "blue", "green"],
            )

    asdf_file.update()



from pathlib import Path

import numpy as np

# TODO replace these with paths from config
PATH = Path.home() / "DKIST" / "vtfcal" / "vtfcal"
DATA = PATH / "data"
TEST_DATA_SIZE = (512, 512)
# TEST_PIXEL = (800, 800)
TEST_PIXEL0 = (125, 50)
rot = 9
c = np.cos(np.deg2rad(rot))
s = np.sin(np.deg2rad(rot))
rm = np.array([[c, -s, 0], [s, c, 0]])
array_centre = np.array(TEST_DATA_SIZE) / 2.0
displacement = np.dot(rm[:2, :2], array_centre) - array_centre
rm[:2, 2] = -displacement
# TEST_PIXEL = np.dot(np.array(TEST_PIXEL0), rmatrix).astype(int)
TEST_PIXEL = (
    int((rm[0, 0] * TEST_PIXEL0[0]) + (rm[0, 1] * TEST_PIXEL0[1]) + rm[0, 2]),
    int((rm[1, 0] * TEST_PIXEL0[0]) + (rm[1, 1] * TEST_PIXEL0[1]) + rm[1, 2]),
)
TEST_WL_IDX = 10
RADIUS = 32



"""
Target frame reduction

Average target frames, correct them for darks and flats, and use them to determine alignment
parameters.

See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

import logging
from os.path import join

import numpy as np
import skimage.feature as skf
import skimage.transform as skt
from skimage.feature import corner_peaks, corner_subpix
from skimage.measure import ransac

import ccdproc

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def average_targets(targetfiles, output_dir):
    """
    Function to reduce VTF target frames

    Loads specified target frames, calculates an average frame per camera channel, corrects these
    for dark and flat effects, then aligns them. Input files should all be for the same line scan.
    Reduced, corrected frames are saved to disk and the corresponding filenames are returned.
    Alignment information is also returned in the form of a rotation matrix.

    \b
    Parameters
    ----------
    targetfiles : list
        List of `pathlib.Path<https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_ objects
        which each point to a FITS file containing a raw VTF flat frame.

    \b
    Returns
    -------
    reduced_flats : list
        List of `pathlib.Path<https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_ objects
        which each point to a FITS file containing a reduced VTF target frame corrected for darks
        and flats.

    rot_matrix : array of floats
        Rotation matrix describing the transformation which aligns each of the narrowband images
        with the broadband channel.

    """

    frames = [ccdproc.CCDData.read(f, format="fits", unit="adu") for f in targetfiles]
    target_fname = join(output_dir, "average_target.FITS")
    target = ccdproc.combine(frames, output_file=target_fname)

    return (target, target_fname)  # latter may not be necessary, can probably get it from metadata


def detect_corners(img):
    return corner_subpix(
        img, corner_peaks(skf.corner_harris(img, 0.2), min_distance=10), window_size=10
    )


def define_alignment_refs(target, extractor):
    points = detect_corners(target)
    extractor.extract(target, points)
    refs = {}
    refs["image"] = target
    refs["points"] = points[extractor.mask]
    refs["descriptors"] = extractor.descriptors

    return points, refs


def align_to_reference(img, extractor, refs):
    # Calculate alignment parameters
    points = detect_corners(img)
    extractor.extract(img, points)
    points = points[extractor.mask]
    descriptors = extractor.descriptors
    matches = skf.match_descriptors(refs["descriptors"], descriptors, cross_check=True)
    transform, inliers = ransac(
        (refs["points"][matches[:, 0]], points[matches[:, 1]]),
        skt.ProjectiveTransform,
        min_samples=5,
        residual_threshold=1,
        max_trials=10000,
    )
    newim = skt.warp(img, transform.inverse, cval=np.nan)
    logger.debug(f"Estimated transform: {transform}")
    logger.debug(f"Estimated transform parameters: {transform.params}")
    logger.debug(f"# inliers: {len(inliers)}")

    return newim, points, matches, transform.params



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This sub-module is destined for common non-package specific utility
# functions.

import configparser as cfp
from pathlib import Path
from .dark_correction_utils import *  # noqa: F401,F403
from .general import *
from .plotting import *

cfg = cfp.ConfigParser()
cfg.read("vtf-calibration-config.cfg")

home = Path.home()
paths = cfg["paths"]
RAW_DATA = Path(paths["raw_data_dir_location"]).expanduser() / paths["raw_data_dir_name"]
DKIST = Path(paths["dkist_dir_location"]).expanduser() / "DKIST"
DATA = DKIST / "VTF" / "data"



import numpy as np


def primes(imax):
    """
    Returns prime numbers up to imax.

    Parameters
    ----------
    imax: int
        The number of primes to return. This should be less or equal to 10000.

    Returns
    -------
    result: list
        The list of prime numbers.
    """

    p = list(range(10000))
    result = []
    k = 0
    n = 2

    if imax > 10000:
        raise ValueError("imax should be <= 10000")

    while len(result) < imax:
        i = 0
        while i < k and n % p[i] != 0:
            i = i + 1
        if i == k:
            p[k] = n
            k = k + 1
            result.append(n)
            if k > 10000:
                break
        n = n + 1

    return result


def do_primes(n, usecython=False):
    if usecython:

        raise Exception("This template does not have the example C code included.")

    else:
        print("Using pure python primes")
        return primes(n)


def main(args=None):

    from astropy.utils.compat import argparse
    from time import time

    parser = argparse.ArgumentParser(description="Process some integers.")
    parser.add_argument(
        "-c",
        "--use-cython",
        dest="cy",
        action="store_true",
        help="Use the Cython-based Prime number generator.",
    )
    parser.add_argument(
        "-t", "--timing", dest="time", action="store_true", help="Time the Fibonacci generator."
    )
    parser.add_argument(
        "-p", "--print", dest="prnt", action="store_true", help="Print all of the Prime numbers."
    )
    parser.add_argument("n", metavar="N", type=int, help="Get Prime numbers up to this number.")

    res = parser.parse_args(args)

    pre = time()
    primes = do_primes(res.n, res.cy)
    post = time()

    print("Found {0} prime numbers".format(len(primes)))
    print("Largest prime: {0}".format(primes[-1]))

    if res.time:
        print("Running time: {0} s".format(post - pre))

    if res.prnt:
        print("Primes: {0}".format(primes))



"""
Polarisation calibration reduction

...

See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

import logging

import asdf

from vtfcal.reduction.polcal import fit_parameters, gen_demod_matrix, gen_mueller_matrix
from vtfcal.utils import correct_darks


def reduce_polcal(data_tree):
    """
    Function to reduce VTF polarisation calibration frames

    Loads polcal frames from the input directory specified by `data_tree` and reduces them by
    applying the following steps:

    - Correct for darks
    - Average accumulations
    - Run reduced polcal frames through DKIST's PA&C pipeline to produce demodulation and telescope
    matrices

    Reduced frames and matrices are saved to the output directory specified by `data_tree` and
    references to the files are added to the tree for use later in the calibration process.

    Parameters
    ----------
    data_tree : string or :class:`pathlib.Path`
        Path to an :class:`~asdf.AsdfFile` defining the calibration data structure, including input
        and output data directories, and file references to polcal frames. See
        :meth:`commands.init_data_tree` for generating an appropriate file.

    Examples
    --------

    """
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    asdf_file = asdf.open(data_tree)
    outdir = asdf_file["support"]["data_dir"]

    # Reduce darks
    # Depends on dark averaging
    # asdf_file = correct_darks(asdf_file, "raw", "polcal")

    # Average accumulations

    # Run through PA&C pipeline
    ## wrapper func around fitcuparams.main()
    fit_parameters(asdf_file)
    ## wrapper func around generatedmodmatrices.main()
    gen_demod_matrix(asdf_file)
    ## wrapper func around telescopemodel.telescopemodel.generate_inverse.telescope_model
    gen_mueller_matrix(asdf_file)



import subprocess as subp

import numpy as np

from astropy.io import fits


def run_testSpeckle(
    width=2048,
    height=2048,
    tile_size=128,
    burst_directory="/home/dkistdev/data",
    burst_prefix="bbso",
    burst_suffix=".pnm",
    burst_size=80,
    fixed_width=0,
    gpus=0,
    master_gpu=0,
    L1=0.949999988,
    L2=0.0900000036,
    iterations=10,
    out_prefix="byproducts",
    fried_parameter=0.100000001,
    wavelength=656.281982,
    pixel_scale=0.0170000009,
    diameter=4,
    dark=None,
    gain=None,
    npsd=None,
    init_with_mean=False,
    floats=False,
):

    # Define bispectrum indices?
    # call spectral ratio library?
    # construct speckle class
    # read darks
    # read flats
    # read 'burst' image files
    # correct for darks and flats
    # sum images?
    # call reconstruction

    if not isinstance(gpus, int):
        gpus = {gpus}

    print(f"{gpus}")

    speckle_command = [
        "./testSpeckle",
        "-w",
        f"{width}",
        "-h",
        f"{height}",
        "-t", f"{tile_size}",
        "-d",
        f"{burst_directory}",
        "-p",
        f"{burst_prefix}",
        "-s",
        f"{burst_suffix}",
        "-n",
        f"{burst_size}",
        f"--fixed-width={fixed_width}",
        f"--gpus={gpus}",
        f"--master-gpu={master_gpu}",
        f"--L1={L1}",
        f"--L2={L2}",
        f"--iterations={iterations}",
        f"--fried-parameter={fried_parameter}",
        f"--wavelength={wavelength}",
        f"--pixel-scale={pixel_scale}",
        f"--diameter={diameter}",
    ]
    if dark:
        speckle_command.append(f"--dark={dark}")
    if gain:
        speckle_command.append(f"--gain={gain}")
    if npsd:
        speckle_command.append(f"--npsd={npsd}")
    if init_with_mean:
        speckle_command.append(f"--init-with-mean={init_with_mean}")
    if floats:
        speckle_command.append(f"--floats={floats}")

    subp.run(speckle_command)

    recon = np.fromfile("byproducts-reconstruction.raw", dtype="float32").reshape(
        (1024, 1024)
    )
    fits.writeto("speckle-output.FITS", recon, overwrite=True)


if __name__ == "__main__":
    run_testSpeckle(
        width=1024,
        height=1024,
        burst_directory="vtfTestImages",
        burst_prefix="cssSim_1024x1024",
        burst_suffix=".raw",
        burst_size=99,
        fixed_width=3,
        gpus=0,
    )






def test_primes():
    from ..example_mod import primes
    assert primes(10) == [2, 3, 5, 7, 11, 13, 17, 19, 23, 29]


def test_deprecation():
    import warnings
    warnings.warn(
        "This is deprecated, but shouldn't raise an exception, unless "
        "enable_deprecations_as_exceptions() called from conftest.py",
        DeprecationWarning)



from .demodulate import demodulate
from .align_frames import align_frames
from .dark_correct import correct_darks
from .flat_correct import correct_flats
from .reduce_darks import reduce_darks
from .reduce_flats import reduce_flats
from .reduce_polcal import reduce_polcal
from .init_data_tree import init_data_tree
from .reconstruct_nb import reconstruct
from .reduce_targets import reduce_targets
from .reconstruct_broadband import reconstruct_broadband



import logging
from os.path import join

import matplotlib.pyplot as plt
import numpy as np

from astropy.io import fits

from vtfcal.test_constants import TEST_PIXEL, TEST_WL_IDX


def plotframes(data_tree, frames, outname, raw=False):
    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")

    nframes = len(frames)
    nrows = np.int(np.ceil(nframes / 4))
    ncols = min(nframes, 4)
    plotshape = (nrows, ncols)
    figsize = (8 * (ncols + 1), 8 * nrows)  # Same for this

    fig, ax = plt.subplots(*plotshape, figsize=figsize)
    try:
        ax = ax.flatten()
    except AttributeError:
        ax = [ax]

    for i, (dkey, title) in enumerate(frames):
        keybase = "raw" if raw else "support"
        logger.debug(f"{keybase} {dkey}")
        files = data_tree[keybase][dkey]
        if len(files) == 1:
            imgfile = files[0].fileuri
        else:
            imgfile = [f.fileuri for f in files if f"l{TEST_WL_IDX:02}" in f.fileuri][0]
        logger.debug(imgfile)

        dat = np.array(fits.open(imgfile)[0].data)
        # vmin = 0 if not (dat < 0).any() else np.nanpercentile(dat, 1)
        vmin = np.nanpercentile(dat, 1)
        vmax = np.nanpercentile(dat, 99)
        if (dat < 0).any() and (dat > 0).any():
            cmap = "coolwarm"
            vext = max(abs(vmax), abs(vmin))
            vmin, vmax = -vext, vext
        else:
            cmap = "magma" if "data" in dkey else "viridis"

        plt.sca(ax[i])
        plt.imshow(dat, cmap=cmap, vmin=vmin, vmax=vmax)
        plt.title(title)
        plt.plot(TEST_PIXEL[0], TEST_PIXEL[1], "x", color="black")

        plt.colorbar()

    plt.savefig(join(data_tree["plots"], outname), bbox_inches="tight")
    plt.close()


def plotprofile(data_tree, frames, outname, raw=False, linestyle=None, color=None):
    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")

    if isinstance(linestyle, str):
        linestyle = [linestyle] * len(frames)
    elif not linestyle:
        linestyle = [None] * len(frames)
    if isinstance(color, str):
        color = [color] * len(frames)
    elif not color:
        color = [None] * len(frames)

    any_norm = any(["normalised" in f[0] for f in frames])
    all_norm = all(["normalised" in f[0] for f in frames])

    fig, ax = plt.subplots(figsize=(16, 8))
    ax1 = ax.twinx() if any_norm and not all_norm else None

    for i, (dkey, title) in enumerate(frames):
        if not isinstance(dkey, str):
            profile = dkey
        else:
            keybase = "raw" if raw else "support"
            logger.debug(f"{keybase} {dkey}")
            files = data_tree[keybase][dkey]
            profile = [
                fits.open(f.fileuri)[0].data[TEST_PIXEL[0], TEST_PIXEL[1]] for f in files
            ]
            is_norm = "normalised" in dkey

        if is_norm and not all_norm:
            plt.sca(ax1)
            plt.axhline(1.0, color='grey', linestyle="--")
        else:
            plt.sca(ax)

        plt.plot(profile, linestyle=linestyle[i], color=color[i], label=title)

    plt.title(f"Profile comparison for pixel {TEST_PIXEL}")
    plt.axvline(TEST_WL_IDX, color="grey", linestyle=":")

    ax.set_xlabel(r"$\lambda$ index")
    ax.set_ylabel(r"$I$")
    fig.legend()

    plt.savefig(join(data_tree["plots"], outname), bbox_inches="tight")
    plt.close()



"""
Polarisation calibration reduction

...

See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

from pathlib import Path

import PAC_Pipeline as pacp

import asdf
from astropy.time import Time

db_file = pacp.generic.get_default_telescope_db()


def fit_parameters(asdf_file):
    """"""
    # get directory of polcal frames and filename for parameter fits from data tree
    tree = asdf_file.tree
    poldir = asdf_file["polarization"]["data_dir"]
    outdir = Path(asdf_file["support"]["data_dir"])
    paramsfile = str(outdir / "polcal-fit-params.FITS")

    # call pa&c fit function
    pacp.FitCUParams.main(poldir, paramsfile, telescope_db=db_file)

    tree["polcal-fit-params"] = paramsfile
    new_asdf = asdf.AsdfFile(tree=tree, uri=asdf_file.uri)
    new_asdf.write_to(asdf_file.uri)


def gen_demod_matrix(asdf_file):
    """"""
    tree = asdf_file.tree
    poldir = Path(asdf_file["polarization"]["data_dir"])
    outdir = Path(asdf_file["support"]["data_dir"])
    paramsfile = asdf_file["polcal-fit-params"]
    demodfile = str(outdir / "polcal-demod-matrix.FITS")

    pacp.GenerateDemodMatrices.main(poldir, paramsfile, demodfile, telescope_db=db_file)

    tree["polcal-demod-matrices"] = demodfile
    new_asdf = asdf.AsdfFile(tree=tree, uri=asdf_file.uri)
    new_asdf.write_to(asdf_file.uri)


def gen_mueller_matrix(asdf_tree):
    """"""
    telescope = pacp.TelescopeModel.TelescopeModel(
        0, 0, 0
    )  ## obviously change these numbers to take from the headers
    obstime = Time("2020-01-31").mjd  # happy unification day
    obswave = 396.86  # nm
    telescope.load_from_database(db_file, obstime, obswave)

    mueller_matrix = telescope.generate_inverse_telescope_model()



"""
Science data preparation

Functions to perform dark corrections on raw science data as part of the preparation for more
rigourous image reconstruction using those data.

See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

import glob
import logging
from os.path import join
from pathlib import Path
from itertools import groupby

import matplotlib.pyplot as plt
import numpy as np

import asdf
from astropy.io import fits

from dkist.asdf_maker import headers_from_filenames as heads
from dkist.asdf_maker import references_from_filenames as refs
from vtfcal import utils
from vtfcal.test_constants import TEST_PIXEL, TEST_WL_IDX


def correct_darks(data_tree):
    """
    Apply dark correction to raw data frames.

    Loads data frames from the input directory specified by `data_tree` and corrects them for dark
    effects using the average dark calculated using :meth:`commands.reduce_darks`. Corrected frames
    are saved to the output directory specified by `data_tree` and references to the files are added
    to the tree for use later in the calibration process.

    \b
    Parameters
    ----------
    data_tree : string or :class:`pathlib.Path`
        Path to an :class:`~asdf.AsdfFile` defining the calibration data structure, including input
        and output data directories, and file references to averaged darks. See
        :meth:`commands.init_data_tree` for generating an appropriate file.

    Examples
    --------

    """
    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")

    asdf_file = asdf.open(data_tree, mode="rw")
    # iterate over groups
    for modstate in ["modstate0", "modstate1", "modstate2", "modstate3"]:
        asdf_file = utils.correct_darks(asdf_file, "raw", "data " + modstate)
        corrected_frames = asdf_file["support"][
            "corrected dark-corrected data " + modstate
        ]
        logger.info(f"Demo plots created with wavelength index {TEST_WL_IDX}")
        logger.debug(corrected_frames)
        # This surely isn't necessary with ExternalArrayReferences ?
        testframe = fits.open(
            [f.fileuri for f in corrected_frames if f"_l{TEST_WL_IDX:02}" in f.fileuri][
                0
            ]
        )[0]

        utils.plotframes(
            asdf_file,
            [
                (
                    f"corrected dark-corrected data {modstate}",
                    "Dark-corrected data frame",
                )
            ],
            "02a-dark-corrected-data",
        )



import logging
import logging.config
from os.path import join, split

import numpy as np

import asdf
import astropy.units as u
import ccdproc

from dkist.asdf_maker import headers_from_filenames as heads
from dkist.asdf_maker import references_from_filenames as refs


def correct_darks(asdf_file, category, data_key):
    """
    Corrects averaged flat frames for telescope dark effects.

    Loads specified dark frames and subtracts those data from the corresponding input flat frames.
    Corrected flats are saved to disk and the files and filenames are returned in a dictionary.

    Parameters
    ----------
    asdf_file : string or `Path<https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_
        Path to an `AsdfFile` defining the calibration data structure, including input and output
        data directories and file references to averaged, dark-corrected flat frames. See
        :meth:`vtf-pipeline.commands.init_data_tree` for generating an appropriate file.

    TODO update this docstring
    tree_keys : list of strings
        Keys specifying what frames are to be corrected. `asdf_file` is accessed with each of the
        keys in turn in a nested way. (see examples).

    Returns
    -------
    new_asdf : :class:`asdf.AsdfFile`
        Updated :class:`asdf.AsdfFile` containing the same information as `data_tree` plus a file
        reference to a map of the wavelength shift value calculated for each pixel.

    Examples
    --------

    """
    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")

    # Load dark files
    ## REFACTOR: file loading
    # this is not robust, it'll break silently as soon as there's >1 dark per channel
    darkfiles = asdf_file["support"]["reduced averaged darks"][0].fileuri
    darkframe = ccdproc.CCDData.read(darkfiles, format="fits", unit="adu")

    files = asdf_file[category][data_key]
    files = sorted([f.fileuri for f in files])
    logger.debug(f"{data_key}, {files}")

    # Establish dark <--> flat correspondence # no need, 1 flat/channel until I work in exp. t's
    # Subtract darks
    corrected_frames = []
    ## OPTIMISE: remove loop
    for f in files:
        path, fname = split(f)
        path = path.replace(asdf_file["raw"]["data_dir"], asdf_file["support"]["data_dir"])
        newpath = join(path, f"dark_corrected_{fname}")
        logger.debug(f"{f} ==> {newpath}")
        data = ccdproc.CCDData.read(f, format="fits", unit="adu")
        logger.debug(
            "++++ averaged: "
            f"{np.nanmin(data.data)} {np.nanmean(data.data)} {np.nanmax(data.data)}"
        )
        # Exposure times here are clearly placeholders and I need to sort it out
        data = ccdproc.subtract_dark(data, darkframe, dark_exposure=1 * u.s, data_exposure=1 * u.s)
        logger.debug(
            "---- corrected: "
            f"{np.nanmin(data.data)} {np.nanmean(data.data)} {np.nanmax(data.data)}"
        )
        data.write(newpath, overwrite=True)
        corrected_frames.append(newpath)

    headers = np.array(heads(corrected_frames))
    old_keys = data_key.replace("reduced ", "").replace("averaged ", "")
    new_key = "corrected dark-corrected " + old_keys
    logger.debug(f"==== {data_key} ==> {new_key}")

    asdf_file["support"][new_key] = refs(corrected_frames, headers, len(corrected_frames))
    asdf_file.update()

    return asdf_file



import configparser as cfp

import matplotlib.pyplot as plt
import numpy as np
from skimage.io import imread
from skimage.transform import AffineTransform, warp

from astropy.io import fits

from vtfcal.utils import RAW_DATA, DATA
from vtfcal.test_constants import TEST_DATA_SIZE

# Set up configuration to be saved out
cfg = cfp.ConfigParser()
cfg["camera-rotations"] = {}
cfg["camera-shifts"] = {}

# Load/create image
raw_image = imread("USAF-1951.png")[..., -1].astype("float64")
# Apply a small scaling across the target image
# This will make it easier to match corners from one image to another
raw_image *= np.linspace(0, 1, raw_image.shape[1])

fig, ax = plt.subplots(1, 3, figsize=(16, 4))
# For each image
for i, channel in enumerate(["broadband", "narrowband1", "narrowband2"]):
    # Apply smallish random rotation and shift
    rot = int((np.random.random() * 30) - 15)
    scale = TEST_DATA_SIZE[0] / np.array(raw_image.shape)  # Assumes square images
    c = np.cos(np.deg2rad(rot))
    s = np.sin(np.deg2rad(rot))
    rmatrix = np.array([[c, -s, 0], [s, c, 0], [0, 0, 1]])
    array_centre = np.array(TEST_DATA_SIZE) / 2.0
    displacement = np.dot(rmatrix[:2, :2], array_centre) - array_centre
    rmatrix[:2, 2] = -displacement
    rmatrix[:2, :2] *= scale
    tform = AffineTransform(matrix=rmatrix)
    new_image = warp(raw_image, tform.inverse, output_shape=TEST_DATA_SIZE, cval=-1)
    if channel == "broadband":
        ref_rot = rot

    # Update config and save out
    cfg["camera-rotations"][channel] = f"{rot:+02}"
    cfg["camera-shifts"][channel] = "0"

    # Save file out
    for m in range(4):
        fits.writeto(RAW_DATA / channel / f"modstate{m}" / "target.FITS", new_image, overwrite=True)
    ax[i].imshow(new_image)
    ax[i].set_title(f"Rotation: {rot:+.3f} (diff: {(ref_rot-rot):+.3f})")

# Save config out
# TODO This needs to interact properly with the cwd config file
with open(RAW_DATA / "calibration-config.cfg", "w") as cfgfile:
    cfg.write(cfgfile)

# Display the misaligned target images
plt.suptitle(f"Target images")

plt.savefig(DATA / "targets.png")
plt.close()



"""
Flat frame reduction

Average collected flat frames as appropriate to reduce to the smallest number required to calibrate
the science data.

See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

import logging
from os.path import join
from pathlib import Path
from multiprocessing.pool import ThreadPool

import dask
import dask.array as da
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
from dask.distributed import Client

import asdf
from astropy.io import fits

from dkist.asdf_maker import headers_from_filenames as heads
from dkist.asdf_maker import references_from_filenames as refs
from dkist.io import DaskFITSArrayContainer as DFAC
from dkist.io.fits import AstropyFITSLoader as Loader
from vtfcal.reduction.flats import average_flats, calculate_wl_shift, correct_wl_shift
from vtfcal.test_constants import TEST_PIXEL, TEST_WL_IDX
from vtfcal.utils import correct_darks, plotframes, plotprofile

# Set up dask for threading of intensive tasks.
# Need this here because Client doesn't like being instantiated not in __main__.
if __name__ == "__main__":
    client = Client()
    dask.config.set(pool=ThreadPool())


def reduce_flats(data_tree, correction=True, fourier=True):
    """
    Function to reduce VTF flat frames

    Loads flat frames from the input directory specified by `data_tree` and reduces them by applying
    the following steps:

    \b
    - Group frames by wavelength position and average them
    - Correct for darks
    - Calculate and correct for wavelength shift
    - Normalise images by scaling each frame by its average value [still needs implementing]

    Reduced flats are saved to the output directory specified by `data_tree` and references to the
    files are added to the tree for use later in the calibration process.

    \b
    Parameters
    ----------
    data_tree : string or :class:`pathlib.Path`
        Path to an :class:`~asdf.AsdfFile` defining the calibration data structure, including input
        and output data directories, and file references to averaged darks. See
        :meth:`commands.init_data_tree` for generating an appropriate file.

    Examples
    --------

    """
    logger = logging.getLogger(__name__)
    logger.setLevel("INFO")

    asdf_file = asdf.open(data_tree, mode="rw")
    outdir = asdf_file["support"]["data_dir"]

    asdf_file = average_flats(asdf_file)
    if asdf_file["mode"] not in ["broadband", "narrowband"]:
        raise ValueError(
            f'Unrecognised data mode in asdf tree: {asdf_file["mode"]} '
            '- "mode" keyword should be either "broadband" or "narrowband".'
        )
    for modstate in ["modstate0", "modstate1", "modstate2", "modstate3"]:
        averaged_frames = asdf_file["support"][
            "reduced averaged flats " + modstate
        ]  # list of reference
        logger.debug(f"{modstate}, {averaged_frames[0]}")

        asdf_file = correct_darks(
            asdf_file, "support", f"reduced averaged flats {modstate}"
        )
        corrected_frames = asdf_file["support"][
            f"corrected dark-corrected flats {modstate}"
        ]

        plotframes(
            asdf_file,
            [
                (f"reduced averaged flats {modstate}", "Averaged flat frames"),
                (f"reduced averaged darks", "Averaged dark frame"),
                (
                    f"corrected dark-corrected flats {modstate}",
                    "Dark-corrected flat frame",
                ),
            ],
            "01aii-flat-correction",
        )

        plotprofile(
            asdf_file,
            [
                (f"reduced averaged flats {modstate}", "Raw flats"),
                (f"corrected dark-corrected flats {modstate}", "Dark-corrected flats"),
            ],
            f"01bii-averaged-vs-corrected-profiles-{modstate}",
            linestyle=[":", "--"],
            color=["red", "blue"],
        )

        if asdf_file["mode"] == "broadband":
            continue

        if not correction:
            arr = DFAC(corrected_frames, loader=Loader).array
            norm_frames = arr / arr.mean(axis=0)
            norm_fnames = []
            for wl, frame in enumerate(norm_frames):
                fname = Path(outdir) / modstate / f"normalised_flat_l{wl:02}a0.FITS"
                fits.writeto(fname, frame.compute(), overwrite=True)
                norm_fnames.append(fname)
            asdf_file["support"]["corrected normalised flats " + modstate] = refs(
                norm_fnames, np.array(heads(norm_fnames)), len(norm_fnames)
            )
            asdf_file.update()

            plotprofile(
                asdf_file,
                [
                    (f"reduced averaged flats {modstate}", "Raw flats"),
                    (
                        f"corrected dark-corrected flats {modstate}",
                        "Dark-corrected flats",
                    ),
                    (mean_profile.reshape(len(norm_frames)), "Average profile"),
                    (f"corrected normalised flats {modstate}", "Normalised flats"),
                ],
                f"profile-comparison-{modstate}",
                linestyle=["-.", ":", "--", None],
                color=["black", "red", "blue", "green"],
            )

            continue

        asdf_file, daskarr, mean_profile = calculate_wl_shift(
            asdf_file, modstate, fourier=fourier
        )

        wl_shift_map = fits.open(
            asdf_file["support"]["calibration wl-shift-map " + modstate][0].fileuri
        )[0].data
        shifted_frames = correct_wl_shift(
            daskarr,
            wl_shift_map,
            mean_profile,
            asdf_file["plots"],
            modstate=modstate,
            fourier=fourier,
        )

        shifted_fnames = []
        for wl, frame in enumerate(shifted_frames):
            fname = Path(outdir) / modstate / f"wl_shifted_flat_l{wl:02}a0.FITS"
            fits.writeto(fname, frame.compute(), overwrite=True)
            shifted_fnames.append(fname)
        asdf_file["support"]["corrected wl-shifted flats " + modstate] = refs(
            shifted_fnames, np.array(heads(shifted_fnames)), len(shifted_fnames)
        )

        plotframes(
            asdf_file,
            [
                (
                    f"corrected dark-corrected flats {modstate}",
                    "Dark-corrected flat frame",
                ),
                (f"calibration wl-shift-map {modstate}", r"$\lambda$-shift map"),
                (
                    f"corrected wl-shifted flats {modstate}",
                    r"$\lambda$-shifted flat frame",
                ),
            ],
            f"01g-wl-shift-comparison",
        )

        logger.debug(shifted_frames.shape)

        norm_frames = (
            shifted_frames
            / da.nanmean(shifted_frames, axis=(1, 2))[:, np.newaxis, np.newaxis]
        )
        norm_fnames = []
        for wl, frame in enumerate(norm_frames):
            fname = Path(outdir) / modstate / f"normalised_flat_l{wl:02}a0.FITS"
            fits.writeto(fname, frame.compute(), overwrite=True)
            norm_fnames.append(fname)
        asdf_file["support"]["corrected normalised flats " + modstate] = refs(
            norm_fnames, np.array(heads(norm_fnames)), len(norm_fnames)
        )
        asdf_file.update()

        plotframes(
            asdf_file,
            [
                (
                    f"corrected wl-shifted flats {modstate}",
                    r"$\lambda$-shifted flat frame",
                ),
                (
                    f"corrected normalised flats {modstate}",
                    "Normalised shifted flat frame",
                ),
            ],
            "01h-normalisation-comparison",
        )

        plotprofile(
            asdf_file,
            [
                (f"corrected wl-shifted flats {modstate}", r"$\lambda$-shifted flats"),
                (mean_profile.reshape(len(norm_fnames)), "Average profile"),
                (f"corrected normalised flats {modstate}", "Normalised flats"),
            ],
            f"01i-normalisation-comparison-{modstate}",
            linestyle=[":", "--", None],
            color=["black", "blue", "green"],
        )



"""
Dark frame reduction

Average collected dark frames as appropriate to reduce to the smallest number required to calibrate
the science data.
"""

import ccdproc


def average_darks(darkfiles, outfile):
    # I feel like there should be a way of loading these all at once.
    # TODO Try just passing multiple filenames and see what it does
    frames = [ccdproc.CCDData.read(f.fileuri, format="fits", unit="adu") for f in darkfiles]

    # Obviously at the moment this just makes one averaged dark
    # Need to code in the unique exposure times part of it
    ## META: exposure times
    ccdproc.combine(frames, output_file=outfile)



import logging
from pathlib import Path

import numpy as np
from skimage.transform import AffineTransform, warp

from astropy.io import fits

from vtfcal.reconstruction.speckle import calc_fried_param, run_speckle


def reconstruct(asdf_file):
    logger = logging.getLogger(__name__)
    # Not sure if I should be doing this one per modstate or not?

    r0 = calc_fried_param()
    # Things that should change here:
    # - autoget width and height (from a previous image I guess, or from run metadata)
    # - separate files into bursts (by timestamps? burst metadata) and repeat below for each burst
    # - autoget burst size from number of files
    # - use burst min/max timestamps to sort AO data for r0 calculation
    specframe = run_speckle(
        width=1024,
        height=1024,
        burst_directory="vtfTestImages",
        burst_prefix="cssSim_1024x1024",
        burst_suffix=".raw",
        burst_size=99,
        fixed_width=3,
        gpus=0,
        fried_parameter=r0,
    )

    outdir = asdf_file["support"]["data_dir"]
    fname = Path(outdir) / f"speckle-output.FITS"

    logger.warning(
        "Speckle images do not match size of simulated VTF data - scaling to match."
    )
    specframe = warp(
        specframe,
        AffineTransform(matrix=np.array([[0.5, 0, 0], [0, 0.5, 0], [0, 0, 1]])).inverse,
        output_shape=(512, 512),
        cval=specframe.mean(),
    )
    fits.writeto(fname, specframe, overwrite=True)

    return asdf_file



import dask.array as da
import dask.bytes
import matplotlib.pyplot as plt
import numpy as np
from dask_kubernetes import KubeCluster

from astropy.io import fits
from astropy.io.fits.hdu.base import BITPIX2DTYPE


def get_info(f):
    with f as fi:
        head = fits.getheader(fi, hdu=1)
    naxes = head["NAXIS"]
    dtype = BITPIX2DTYPE[head["BITPIX"]]
    shape = [head[f"NAXIS{n}"] for n in range(naxes, 0, -1)]
    return dtype, shape


class DelayedFits:
    def __init__(self, file, shape, dtype):
        self.shape = shape
        self.dtype = dtype
        self.file = file
        self.hdu = 0

    def __getitem__(self, item):
        with self.file as f:
            with self.file as f:
                with fits.open(f) as hdul:
                    hdul.verify("fix")
                    return hdul[self.hdu].data[item]


files = dask.bytes.open_files(
    "/home/drew/vtfcal/vtfcal/data/test_input/narrowband_t/modstate0/flat_*.fits.gz"
)
a = list(map(get_info, files))
for f in files:
    print(f.path)
n = np.array(a, dtype="object")
assert all([n[0, 0] == a for a in n[:, 0]])
assert all([n[0, 1] == a for a in n[:, 1]])
dtype = n[0, 0]
shape = n[0, 1]
arrs = [da.from_array(DelayedFits(fn, shape=shape, dtype=dtype), chunks=shape) for fn in files]
arr = da.stack(arrs)
print(arr)
plt.plot(arr[:, 512, 512].compute())
plt.show()



"""
Image reconstruction


See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

import logging
from os.path import join
from pathlib import Path

import dask.array as da
import matplotlib.pyplot as plt
import numpy as np
from numpy.fft import fft2, fftshift, ifft2, ifftshift

import asdf

from dkist.asdf_maker import headers_from_filenames as heads
from dkist.asdf_maker import references_from_filenames as refs
from dkist.io import DaskFITSArrayContainer as DFAC
from dkist.io.fits import AstropyFITSLoader as Loader
from vtfcal.test_constants import TEST_WL_IDX
from vtfcal.utils import correct_darks


def apod_filter(image, width):
    x, y = image.shape[-2:]
    xx, yy = np.mgrid[:x, :y]
    r = np.hypot(xx, yy)
    apod = 0.5 * (1 - np.cos(np.pi * r / width))
    apod[width:-width, width:-width] = 1

    return image * apod


def fft_chunks(image):
    return fftshift(fft2(apod_filter(image, 8)))
    # return fft2(image)


def calc_noise(fft_bb, fsum, bsum):
    return np.abs(fft_bb * (fsum / bsum)) ** 2


def calc_im(fft_bb, nsum, bsum):
    # Really not sure if this should be fft_bb or fftspeck_bb, but I'm doing this for now to make it
    # actually run
    # recon_im = fftspeck_bb * (nsum / bsum)
    return fft_bb * (nsum / bsum)


def calc_power(recon_im):
    # Summing actually breaks the algorithm and I need to figure out why and what to do about it
    return (np.abs(recon_im) ** 2)  #.sum()


def calc_optfilter(recon_noise, data_power):
    opt_filter = (data_power - recon_noise) / data_power
    # Shift interesting parts of the spectrum to the centre
    # opt_filter = fftshift(opt_filter)
    # smooth optimum_filter

    return opt_filter


def reconstruct_image(recon_im, opt_filter):
    filtered_im = recon_im * opt_filter
    recon_nb = ifft2(ifftshift(filtered_im))
    # recon_nb = ifft2(filtered_im)

    return recon_nb


def reconstruct(data_tree, bb_tree=None):
    """
    ;; Read data and reconstruction
    FOR each (wavelength?) step st
        ;; Average the linear fits for each subimage
        FOR each subimage sub
            # estimate noise in reconstruction
            ;; Result is the fft of the reconstructed data.  We cannot
            ;; transform back yet, the optimum filter must be created
            ;; first.  Therefore store the FFTs
        ## compute the optimum filter:
        # First, sum the power of the reconstructed data
        # Center the filter, the interesting parts are at the border 
        ;; There are still a lot of peaks at high frequencies.  Use an
        ;; filter to mask them out.
        ;; Using a symmetric filter is a good idea:  Do a polar
        ;; integration and find (by interpolation) the value where it 
        ;; drops below .4 Create a fermi-type filter with that cutoff
        ;; Decenter the optimum filter back to fourier-representation
        ;; Finish reconstruction: apply optimum filter, transform back to data
        FOR each subimage sub
            # FFT back
            ;; Compute so far connected image
            ;; Total intensity in each part
            ;; Correction factor for intensity

    for each state in polarimetric cycle:
        bb = read broadband images
        bb_rec = read speckle-reconstructed broadband image
        bb_rec` = calculate FFT of bb_rec
        nb, flat = read narrowband images, flats

        for each step in line scan:

            for each patch in image:

                for each acquisition:
                    bb`, nb`, flat` = calculate FFT of bb[patch, acquisition], nb[patch, acquisition], flat[patch, acquisition]
                    bsum[patch, acquisition], nsum[patch, acquisition], fsum[patch, acquisition] = real((bb`, nb`, flat`) * conj(bb`))

                sum each of bsum, nsum, fsum over acquisitions
                reconstruction_noise[patch] = power(bb`[patch] * (fsum / bsum))
                unfiltered reconstructed image nb_rec``[patch] = bb_rec` * (nsum / bsum)

            data_power = sum(power(nb_rec``[patch]) for each patch)
            optimum_filter = (data_power - reconstruction_noise) / data_power
            smooth optimum_filter

            for each patch in image:
                filtered reconstructed image in Fourier domain nb_rec`[patch] = nb_rec`` * optimum_filter
                reconstructed data nb_rec[patch] = IFFT(nb_rec`[patch])

            output nb_rec[state, step] = recombine and normalise (nb_rec[patch] for each patch)

    Function to reconstruct VTF science data

    Loads science frames from the input directory specified by `data_tree` and reconstructs them by
    applying the process described in the VTF algorithm document [1]_.

    Parameters
    ----------
    data_trees : string or :class:`pathlib.Path`
        Paths to :class:`~asdf.AsdfFile`s defining the calibration data structure, including input and output
        data directories, and file references to prepared science data. See
        :meth:`commands.init_data_tree` for generating appropriate files.

    Examples
    --------

    References
    ----------
    .. _[1] VTF/DKIST data pipeline - The algorithm structure
    """
    logger = logging.getLogger(__name__)

    asdf_file = asdf.open(data_tree, mode="rw")
    bb_tree = asdf.open(bb_tree, mode="rw")
    if asdf_file["mode"] == "narrowband" and not bb_tree:
        raise ValueError(
            "Location of broadband data structure must be provided for reconstruction of narrowband data."
        )
    elif asdf_file["mode"] != "narrowband":
        raise ValueError(
            f'Unrecognised data mode in asdf tree: {asdf_file["mode"]} '
            '- "mode" keyword should be either "broadband" or "narrowband".'
        )

    # Dark-correct speckle calibration files
    # I assume I need this but I'm not sure if I need to do flats as well
    speckfile = Path(bb_tree["support"]["data_dir"]) / "speckle-output.FITS"
    headers = np.array(heads([speckfile]))
    bb_tree["support"]["speckle"] = refs(speckfile, headers, 1)
    bb_tree.update()
    asdf_file = correct_darks(bb_tree, "support", "speckle")

    # polstate = get_polarisation_state()

    # These also obviously need changing to work for all modstates
    nb = DFAC(
        asdf_file["support"]["corrected flat-corrected data modstate0"], loader=Loader
    ).array
    bb = DFAC(
        bb_tree["support"]["corrected flat-corrected data modstate0"], loader=Loader
    ).array
    # flat = DFAC(asdf_file["support"]["corrected wl_shifted flats"], loader=Loader).array
    flat = DFAC(
        asdf_file["support"]["corrected dark-corrected flats modstate0"], loader=Loader
    ).array
    # This likely won't be where the speckle files end up
    logger.debug(bb_tree["support"].keys())
    speckled_bb = DFAC(
        bb_tree["support"]["corrected dark-corrected speckle"], loader=Loader
    ).array
    chunksize = (32, 32)
    overlap = {0: 0, 1: 8, 2: 8}
    nb = da.rechunk(nb, (1, *chunksize))
    bb = da.rechunk(bb, (1, *chunksize))
    flat = da.rechunk(flat, (1, *chunksize))
    speckled_bb = da.rechunk(speckled_bb, chunksize)
    fftspeck_bb = speckled_bb.map_overlap(fft_chunks, depth=(overlap[1], overlap[2]))
    fftspeck_bb = da.rechunk(
        fftspeck_bb.reshape(1, *fftspeck_bb.shape), (1, *chunksize)
    )

    plt.close("all")
    # for each step in line scan:
    # for each patch in image:
    #     for each acquisition:
    #         bb`, nb`, flat` = calculate FFT of bb[patch, acquisition], nb[patch, acquisition], flat[patch, acquisition]
    #         bsum[patch, acquisition], nsum[patch, acquisition], fsum[patch, acquisition] = real((bb`, nb`, flat`) * conj(bb`))
    #     sum each of bsum, nsum, fsum over acquisitions
    #     reconstruction_noise[patch] = power(bb`[patch] * (fsum / bsum))
    #     unfiltered reconstructed image nb_rec``[patch] = bb_rec` * (nsum / bsum)

    fft_bb = bb.map_overlap(fft_chunks, depth=overlap)
    fft_nb = nb.map_overlap(fft_chunks, depth=overlap)
    fft_flat = flat.map_overlap(fft_chunks, depth=overlap)

    conjugate = fft_bb.conj()
    bsum = (fft_bb * conjugate).real
    nsum = (fft_nb * conjugate).real
    fsum = (fft_flat * conjugate).real
    # sum each of bsum, nsum, fsum over acquisitions

    recon_noise = da.map_overlap(calc_noise, fft_bb, fsum, bsum, depth=overlap)
    recon_im = da.map_overlap(calc_im, fftspeck_bb, nsum, bsum, depth=overlap)
    data_power = recon_im.map_overlap(calc_power, depth=overlap)

    # data_power = sum(power(nb_rec``[patch]) for each patch)
    # optimum_filter = (data_power - reconstruction_noise) / data_power

    # smooth optimum_filter

    optimum_filter = da.map_overlap(
        calc_optfilter, recon_noise, data_power, depth=overlap
    )

    # for each patch in image:
    #     filtered reconstructed image in Fourier domain nb_rec`[patch] = nb_rec`` * optimum_filter
    #     reconstructed data nb_rec[patch] = IFFT(nb_rec`[patch])

    # output nb_rec[state, step] = recombine and normalise (nb_rec[patch] for each patch)

    recon_nb = da.map_overlap(
        reconstruct_image, recon_im, optimum_filter, depth=overlap
    ).compute()
    img = recon_nb[TEST_WL_IDX].real
    logger.debug(f"{img.min()}, {img.mean()}, {img.max()}, {img.std()}")

    logger.debug(f"{nb.shape}, {recon_nb.shape}")
    fig, ax = plt.subplots(1, 2, figsize=(12, 4))
    ax[0].imshow(nb[TEST_WL_IDX].compute())
    vmin = img.mean() - img.std()
    vmax = img.mean() + img.std()
    ax[1].imshow(img)  # , vmin=vmin, vmax=vmax)
    plt.savefig(Path(asdf_file["plots"]) / "reconstructed-narrowband")
    plt.close()



"""
"""

import logging
from pathlib import Path
import asdf
from vtfcal.reconstruction.broadband import reconstruct

def reconstruct_broadband(data_tree):
    """
    """

    logger = logging.getLogger(__name__)
    asdf_file = asdf.open(data_tree, mode="rw")

    if asdf_file["mode"] != "broadband":
        raise ValueError(
            f'Invalid data mode keyword in asdf tree: {asdf_file["mode"]} '
            'Speckle reconstruction is only applicable to data with "mode": "broadband"')

    asdf_file = reconstruct(asdf_file)

    asdf_file.close()



import logging
from vtfcal.utils import correct_darks

def reduce_prefilter(data_tree):
    """
    """

    logger = logging.getLogger(__name__)

    asdf_file = asdf.open(data_tree, mode="rw")

    # Dark correct prefilter frames
    asdf_file = correct_darks(asdf_file, "raw", "prefilter")
    corrected_frames = asdf_file["support"]["corrected dark-corrected prefilter"]
    # Average across accumulations
    # Correct wavelength shift in prefilter frames
    # Build scan curve - average over x, y for each frame to just leave lambda
    # Align FTS atlas spectrum with scan curve
    # Divide scan by FTS spectrum
    # Fit double-Gaussian



"""
Dark frame reduction

Average collected dark frames as appropriate to reduce to the smallest number required to calibrate
the science data.

See Also
--------
Optional

Notes
-----
Optional

References
----------
Optional, use if references are cited in Notes

Examples
--------
Optional
"""

from os.path import join

import asdf

from vtfcal.utils import average_all

PATH = join("/", "home", "drew", "DKIST" "vtfcal", "vtfcal")


## TODO Change this and other functions to take a FOLDER, not a file, and to look there for an appropriate asdf file
## Will mean most functions needing to take a channel keyword also
## Possibly specify default asdf file location in a config file?
## Also most of the content of this function should be moved to one in the vtfcal package
## More generally, these command-line functions should do as little as possible,
## other than calling the appropriate vtfcal function and minor asdf file-handling
def reduce_darks(data_tree):
    """
    Function to reduce VTF dark frames

    Loads all available dark frames from the input directory specified by `data_tree`. These are
    averaged into a single frame, which is saved to the appropriate output directory. A reference to
    the averaged dark file is also added to `data_tree` for use later in the calibration process.

    \b
    Parameters
    ----------
    data_tree : string or :class:`pathlib.Path`
        Path to an :class:`~asdf.AsdfFile` defining the input and output data directories for the
        calibration, and whether the data are broadband or narrowband images. See
        :meth:`commands.init_data_tree` for generating an appropriate file.
    """

    tree = asdf.open(data_tree, mode="rw")

    ## REFACTOR: This will average broadband darks by wavelength as well which I don't want
    average_all(tree, "darks")



import glob

from astropy.io import fits

from dkist_data_model.generator import VTF

vtf = VTF(end_condition="1s")
h, c = list(vtf.generate_metadata("mode1"))[0]

for fname in glob.iglob("data/test_input/**/*.FITS", recursive=True):
    f = fits.open(fname)[0]
    for k, v in h.items():
        f.header[k] = v
        if k in list(c.keys()):
            f.header.comments[k] = c[k]
    # f.header.set('DNAXIS', 3)
    # f.header.set('DNAXIS1', 512)
    # f.header.set('DNAXIS2', 512)
    # f.header.set('DNAXIS3', 18)
    # f.header.set('DAAXES', 2)
    # f.header.set('DTYPE1', h['CTYPE1'])
    # f.header.set('DTYPE2', h['CTYPE2'])
    # f.header.set('DTYPE3', 'SPECTRAL')
    # f.header.set('DUNIT1', h['CUNIT1'])
    # f.header.set('DUNIT2', h['CUNIT2'])
    # f.header.set('DUNIT3', 'nm')
    # f.header.set('DPNAME1', 'spatial x')
    # f.header.set('DPNAME2', 'spatial y')
    # f.header.set('DPNAME3', 'lambda')
    # f.header.set('DWNAME1', 'helioprojetive latitude')
    # f.header.set('DWNAME2', 'helioprojective longitude')
    # f.header.set('DWNAME3', 'wavelength')
    # f.header.set('DINDEX3', 0)

    f.writeto(fname, overwrite=True)
